from enum import IntEnum
from . import kind_helper


class DataPointLabelPlacementKind(IntEnum):
    """Data Point Label Placement Kind"""

    ABOVE = 0
    CENTER = 1
    LEFT = 4
    BELOW = 6
    RIGHT = 8

    @staticmethod
    def from_str(s: str) -> "DataPointLabelPlacementKind":
        """
        Gets an ``DataPointLabelPlacementKind`` instance from string.

        Args:
            s (str): String that represents the name of an enum Name.
                ``s`` is case insensitive and can be ``CamelCase``, ``pascal_case`` , ``snake_case``,
                ``hypen-case``, ``normal case``.

        Raises:
            ValueError: If input string is empty.
            AttributeError: If unable to get ``DataPointLabelPlacementKind`` instance.

        Returns:
            DataPointLabelPlacementKind: Enum instance.
        """
        return kind_helper.enum_from_string(s, DataPointLabelPlacementKind)



# coding: utf-8
from __future__ import annotations
from typing import List
from pathlib import Path
import uuid
from abc import ABC, abstractmethod

from ..utils import paths


class ConnectorBase(ABC):
    @abstractmethod
    def get_connnection_str(self) -> str:
        ...


class ConnectorBridgeBase(ConnectorBase):
    def __init__(self, **kwargs) -> None:
        self._no_restore = bool(kwargs.get("no_restore", True))
        self._no_first_start_wizard = bool(kwargs.get("no_first_start_wizard", True))
        self._no_logo = bool(kwargs.get("no_logo", True))
        self._invisible = bool(kwargs.get("invisible", True))
        self._headless = bool(kwargs.get("headless", False))
        self._start_as_service = bool(kwargs.get("start_as_service", False))
        self._start_office = bool(kwargs.get("start_office", True))

        if "soffice" in kwargs:
            self.soffice = kwargs["soffice"]

    def update_startup_args(self, args: List[str]) -> None:
        if self.no_restore:
            args.append("--norestore")
        if self.invisible:
            args.append("--invisible")
        if self.no_restore:
            args.append("--norestore")
        if self.no_first_start_wizard:
            args.append("--nofirststartwizard")
        if self.no_logo:
            args.append("--nologo")
        if self.headless:
            args.append("--headless")

    @property
    def soffice(self) -> Path:
        """
        Get/Sets Path to LibreOffice soffice. Default is auto discovered.
        """
        try:
            return self._soffice
        except AttributeError:
            self._soffice = paths.get_soffice_path()
        return self._soffice

    # region startup flags
    @property
    def start_office(self) -> bool:
        """Gets/Sets if office is to be started. Default is True"""
        return self._start_office

    @start_office.setter
    def start_office(self, value: bool):
        self._start_office = value

    @property
    def no_restore(self) -> bool:
        """Gets/Sets if office is started with norestore Option. Default is True"""
        return self._no_restore

    @no_restore.setter
    def no_restore(self, value: bool):
        self._no_restore = value

    @property
    def no_first_start_wizard(self) -> bool:
        """Gets/Sets if office is started with nofirststartwizard option. Default is True"""
        return self._no_first_start_wizard

    @no_first_start_wizard.setter
    def no_first_start_wizard(self, value: bool):
        self._no_first_start_wizard = value

    @property
    def no_logo(self) -> bool:
        """Gets/Sets if office is started with nologo option. Default is True"""
        return self._no_logo

    @no_logo.setter
    def no_logo(self, value: bool):
        self._no_logo = value

    @property
    def invisible(self) -> bool:
        """Gets/Sets if office is started with invisible option. Default is True"""
        return self._invisible

    @invisible.setter
    def invisible(self, value: bool):
        self._invisible = value

    @property
    def headless(self) -> bool:
        """Gets/Sets if the connection is made using headless option. Default is False"""
        return self._headless

    @headless.setter
    def headless(self, value: bool):
        self._headless = value

    @property
    def start_as_service(self) -> bool:
        """
        Gets/Sets if office is started as service (StarOffice.Service).
        Default is False
        """
        return self._start_as_service

    @start_as_service.setter
    def start_as_service(self, value: bool):
        self._start_as_service = value

    # endregion startup flags


class ConnectSocket(ConnectorBridgeBase):
    """Connect to LO via socket"""

    def __init__(self, host="localhost", port=2002, **kwargs) -> None:
        """
        Constructor

        Args:
            host (str, optional): Connection host. Defaults to ``localhost``.
            port (int, optional): Connection port. Defaults to ``2002``.

        Keyword Arguments:
            no_restore (bool, optional): Default ``True``
            no_first_start_wizard (bool, optional): Default ``True``
            no_logo (bool, optional): Default ``True``
            invisible (bool, optional): Default ``True``
            headless (bool, optional): Default ``False``
            start_as_service (bool, optional): Default ``False``
            start_office (bool, optional): Default ``True``
        """
        super().__init__(**kwargs)
        self._host = host
        self._port = port

    def get_connnection_str(self) -> str:
        identifier = f"socket,host={self.host},port={self.port}"
        conn_str = f"uno:{identifier};urp;StarOffice.ServiceManager"
        return conn_str

    @property
    def host(self) -> str:
        """
        Gets/Sets host. Default ``localhost``
        """
        return self._host

    @host.setter
    def host(self, value: str):
        self._host = value

    @property
    def port(self) -> int:
        """
        Gets/Sets port. Default ``2002``
        """
        return self._port

    @port.setter
    def port(self, value: int):
        self._port = value


class ConnectPipe(ConnectorBridgeBase):
    """Connect to LO via pipe"""

    def __init__(self, pipe: str | None = None, **kwargs) -> None:
        """
        Constructor

        Args:
            pipe (str | None, optional): Name of pipe. Auto generated if None. Defaults to ``None``.

        Keyword Arguments:
            no_restore (bool, optional): Default ``True``
            no_first_start_wizard (bool, optional): Default ``True``
            no_logo (bool, optional): Default ``True``
            invisible (bool, optional): Default ``True``
            headless (bool, optional): Default ``False``
            start_as_service (bool, optional): Default ``False``
            start_office (bool, optional): Default ``True``
        """
        super().__init__(**kwargs)
        if pipe is None:
            self._pipe = uuid.uuid4().hex
        else:
            self._pipe = pipe

    def get_connnection_str(self) -> str:
        identifier = f"pipe,name={self.pipe}"
        conn_str = f"uno:{identifier};urp;StarOffice.ServiceManager"
        return conn_str

    @property
    def pipe(self) -> str:
        """
        Gets/Sets pipe used to connect. Default is auto generated hex value
        """
        return self._pipe

    @pipe.setter
    def pipe(self, value: str):
        self._pipe = value



# coding: utf-8
from __future__ import annotations
import io
import base64
from typing import TYPE_CHECKING
from PIL import Image  # LibreOffice has PHL Module.
import uno

if TYPE_CHECKING:
    from com.sun.star.graphic import XGraphic

from ..utils import file_io as mFileIO
from . import images_lo as mImgLo

from ..utils.type_var import PathOrStr


class Images(mImgLo.ImagesLo):
    @staticmethod
    def load_image(fnm: PathOrStr) -> Image.Image:
        img = None
        try:
            img = Image.open(fnm)
            img.load()
            print(f"Loaded image: '{fnm}'")
        except Exception as e:
            print(f"Unable to load '{fnm}':")
            print(f"    {e}")
        return img

    @staticmethod
    def save_image(im: Image.Image, fnm: PathOrStr) -> None:
        """
        Saves Image

        Args:
            im (Image): Image data
            fnm (PathOrStr): save path

        Raises:
            TypeError: If im is None
            Exception: If unable to save image
        """
        if im is None:
            raise TypeError("im is None")
        try:
            im.save(fp=fnm, format="png")
            print(f"Saved image to file: {fnm}")
        except Exception as e:
            raise Exception(f"Could not save image to '{fnm}'") from e

    @staticmethod
    def im_to_bytes(im: Image.Image) -> bytes:
        """
        Gets Image Bytes

        Args:
            im (Image): Image

        Returns:
            bytes: Image bytes
        """
        buf = io.BytesIO()
        im.save(buf, format="png")
        byte_im = buf.getvalue()
        return byte_im

    @classmethod
    def im_to_string(cls, im: Image.Image) -> str:
        """
        Converts image to string

        Args:
            im (Image): image

        Raises:
            Exception: If unable to convet to string

        Returns:
            str: Image as string
        """
        try:
            b = cls.im_to_bytes(im)
            s = base64.b64encode(b).decode("utf-8")
            return s
        except Exception as e:
            raise Exception("Converting image to string is not possible:") from e

    @staticmethod
    def string_to_im(s: str) -> Image.Image:
        try:
            imgdata = base64.b64decode(s)
            image = Image.open(io.BytesIO(imgdata))
            return image
        except Exception as e:
            print("Converting string to image is not possible:")
            print(f"    {e}")
            return None

    @staticmethod
    def bytes_to_im(b: bytes) -> Image.Image:
        try:
            image = Image.open(io.BytesIO(b))
            return image
        except Exception as e:
            print("Converting bytes to image is not possible:")
            print(f"    {e}")
            return None

    @classmethod
    def im_to_graphic(cls, im: Image.Image) -> XGraphic | None:
        if im is None:
            print("No image found")
            return None

        temp_fnm = mFileIO.FileIO.create_temp_file(im_format="png")
        if temp_fnm is None:
            print("Could not create a temporary file for the image")
            return None

        im.save(temp_fnm, format="png")
        graphic = cls.load_graphic_file(temp_fnm)
        mFileIO.FileIO.delete_file(temp_fnm)
        return graphic

    @classmethod
    def graphic_to_im(cls, graphic: XGraphic) -> Image.Image | None:
        if graphic is None:
            print("No graphic found")
            return None
        tmp_fnm = mFileIO.FileIO.create_temp_file("png")
        if tmp_fnm is None:
            return None
        cls.save_graphic(graphic, tmp_fnm, "png")
        im = cls.load_image(tmp_fnm)
        mFileIO.FileIO.delete_file(tmp_fnm)
        return im



# coding: utf-8
"""
Internal Module only! DO NOT use this module/class!

This module is for the purpose of sharing events between classes internally
"""
from __future__ import annotations
from weakref import ref, ReferenceType
from .args.event_args import EventArgs
from typing import List, Dict
from ..utils import type_var
from ..proto import event_observer


class _Events(object):
    """
    Singleton Class for sharing events among internal classes. DO NOT USE!

    Use: lo_events.LoEvents for global events. Use lo_events.Events for locally scoped events.
    """

    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(_Events, cls).__new__(cls, *args, **kwargs)
            cls._instance._callbacks = None
            cls._instance._observers: List[ReferenceType[event_observer.EventObserver]] = None
        return cls._instance

    def on(self, event_name: str, callback: type_var.EventCallback):
        """
        Registers an event

        Args:
            event_name (str): Unique event name
            callback (Callable[[object, EventArgs], None]): Callback function
        """
        if self._callbacks is None:
            self._callbacks: Dict[str, List[ReferenceType[type_var.EventCallback]]] = {}

        if event_name not in self._callbacks:
            self._callbacks[event_name] = [ref(callback)]
        else:
            self._callbacks[event_name].append(ref(callback))

    def trigger(self, event_name: str, event_args: EventArgs, *args, **kwargs) -> None:
        """
        Trigger event(s) for a given name.

        Args:
            event_name (str): Name of event to trigger/
            event_args (EventArgs): Event args passed to the callback for trigger.
            args (Any, optional): Optional positional args to pass to callback
            kwargs (Any, optional): Optional keyword args to pass to callback
        """
        if self._callbacks is not None and event_name in self._callbacks:
            cleanup = None
            for i, callback in enumerate(self._callbacks[event_name]):
                if callback() is None:
                    if cleanup is None:
                        cleanup = []
                    cleanup.append(i)
                    continue
                if event_args is not None:
                    event_args._event_name = event_name
                    if event_args.event_source is None:
                        event_args._event_source = self
                if callable(callback()):
                    try:
                        callback()(event_args.source, event_args, *args, **kwargs)
                    except AttributeError:
                        # event_arg is None
                        callback()(self, None)
            if cleanup is not None and len(cleanup) > 0:
                # reverse list to allow removing form highest to lowest to avoid errors
                cleanup.reverse()
                for i in cleanup:
                    self._callbacks[event_name].pop(i)
                if len(self._callbacks[event_name]) == 0:
                    del self._callbacks[event_name]
        self._update_observers(event_name, event_args)

    def _update_observers(self, event_name: str, event_args: EventArgs) -> None:
        if self._observers is not None:
            cleanup = None
            for i, observer in enumerate(self._observers):
                if observer() is None:
                    if cleanup is None:
                        cleanup = []
                    cleanup.append(i)
                    continue
                observer().trigger(event_name=event_name, event_args=event_args)
            if cleanup is not None and len(cleanup) > 0:
                # reverse list to allow removing form highest to lowest to avoid errors
                cleanup.reverse()
                for i in cleanup:
                    self._observers.pop(i)

    def add_observer(self, *args: event_observer.EventObserver) -> None:
        """
        Adds observers that gets their ``trigger`` method called when this class ``trigger`` method is called.
        """
        if self._observers is None:
            self._observers = []
        for observer in args:
            self._observers.append(ref(observer))



# coding: utf-8
"""
Global Named Events.
"""
from __future__ import annotations
from typing import NamedTuple


class GblNamedEvent(NamedTuple):
    """
    Global Named Events
    """

    PRINTING = "global_printing"
    """Global Printing Event."""



from enum import IntEnum
from . import kind_helper


class AxisKind(IntEnum):
    X = 0
    """Represents X Axis"""
    Y = 1
    """Represents Y Axis"""
    Z = 2
    """Represents Z Axis"""

    @staticmethod
    def from_str(s: str) -> "AxisKind":
        """
        Gets an ``AxisKind`` instance from string.

        Args:
            s (str): String that represents the name of an enum Name.

        Raises:
            ValueError: If input string is empty.
            AttributeError: If unable to get ``AxisKind`` instance.

        Returns:
            AxisKind: Enum instance.
        """
        return kind_helper.enum_from_string(s, AxisKind)



# region Imports
from __future__ import annotations
from random import random
from typing import List, Tuple, cast, overload, TYPE_CHECKING

import uno

# XChartTypeTemplate import error in LO >+ 7.4
from com.sun.star.chart2 import XChartTypeTemplate
from com.sun.star.beans import XPropertySet
from com.sun.star.chart2 import XAxis
from com.sun.star.chart2 import XChartDocument
from com.sun.star.chart2 import XChartType
from com.sun.star.chart2 import XChartTypeContainer
from com.sun.star.chart2 import XCoordinateSystem
from com.sun.star.chart2 import XCoordinateSystemContainer
from com.sun.star.chart2 import XDataSeries
from com.sun.star.chart2 import XDataSeriesContainer
from com.sun.star.chart2 import XDiagram
from com.sun.star.chart2 import XFormattedString
from com.sun.star.chart2 import XLegend
from com.sun.star.chart2 import XRegressionCurve
from com.sun.star.chart2 import XRegressionCurveContainer
from com.sun.star.chart2 import XScaling
from com.sun.star.chart2 import XTitle
from com.sun.star.chart2 import XTitled
from com.sun.star.chart2.data import XDataProvider
from com.sun.star.chart2.data import XDataSink
from com.sun.star.chart2.data import XDataSource
from com.sun.star.chart2.data import XLabeledDataSequence
from com.sun.star.container import XNameAccess
from com.sun.star.document import XEmbeddedObjectSupplier
from com.sun.star.drawing import XDrawPage
from com.sun.star.drawing import XDrawPageSupplier
from com.sun.star.drawing import XShape
from com.sun.star.embed import XComponentSupplier
from com.sun.star.embed import XEmbeddedObject
from com.sun.star.graphic import XGraphic
from com.sun.star.lang import XComponent
from com.sun.star.lang import XMultiServiceFactory
from com.sun.star.sheet import XSpreadsheet
from com.sun.star.sheet import XSpreadsheetDocument
from com.sun.star.table import XTableChart
from com.sun.star.table import XTableChartsSupplier
from com.sun.star.util import XNumberFormatsSupplier

from . import calc as mCalc
from ..exceptions import ex as mEx
from ..utils import color as mColor
from ..utils import file_io as mFileIo
from ..utils import gui as mGui
from ..utils import images_lo as mImgLo
from ..utils import info as mInfo
from ..utils import lo as mLo
from ..utils import props as mProps
from ..utils.data_type.angle import Angle as Angle
from ..utils.kind.axis_kind import AxisKind as AxisKind
from ..utils.kind.chart2_data_role_kind import DataRoleKind as DataRoleKind
from ..utils.kind.chart2_types import ChartTemplateBase, ChartTypeNameBase, ChartTypes as ChartTypes
from ..utils.kind.curve_kind import CurveKind as CurveKind
from ..utils.kind.data_point_label_type_kind import DataPointLabelTypeKind as DataPointLabelTypeKind
from ..utils.kind.line_style_name_kind import LineStyleNameKind as LineStyleNameKind

from ooo.dyn.awt.rectangle import Rectangle
from ooo.dyn.chart.chart_data_row_source import ChartDataRowSource
from ooo.dyn.chart.error_bar_style import ErrorBarStyle
from ooo.dyn.chart2.axis_orientation import AxisOrientation
from ooo.dyn.chart2.axis_type import AxisTypeEnum as AxisTypeKind
from ooo.dyn.chart2.data_point_geometry3_d import DataPointGeometry3DEnum as DataPointGeometry3DEnum
from ooo.dyn.chart2.data_point_label import DataPointLabel
from ooo.dyn.drawing.fill_style import FillStyle as FillStyle
from ooo.dyn.drawing.line_style import LineStyle as LineStyle
from ooo.dyn.lang.locale import Locale
from ooo.dyn.table.cell_range_address import CellRangeAddress

if TYPE_CHECKING:
    from ooo.lo.chart2.data_point_properties import DataPointProperties

# endregion Imports


class Chart2:
    """Chart 2 Class"""

    _CHART_NAME = "chart$$_"

    # region insert a chart
    @classmethod
    def insert_chart(
        cls,
        sheet: XSpreadsheet,
        cells_range: CellRangeAddress,
        cell_name: str,
        width: int,
        height: int,
        diagram_name: ChartTemplateBase | str,
        color_bg: mColor.Color = mColor.CommonColor.PALE_BLUE,
        color_wall: mColor.Color = mColor.CommonColor.LIGHT_BLUE,
    ) -> XChartDocument:
        """
        Insert a new chart

        Args:
            sheet (XSpreadsheet): Spreadsheet
            cells_range (CellRangeAddress): Cell range address
            cell_name (str): Cell name such as ``A1``
            width (int): Width
            height (int): Height
            diagram_name (ChartTemplateBase | str): Diagram Name
            color_bg (Color, optional): Color Background. Defaults to ``CommonColor.PALE_BLUE``.
            color_wall (Color, optional): Color Wall. Defaults to ``CommonColor.LIGHT_BLUE``.

        Raises:
            ChartError: If error occurs

        Returns:
            XChartDocument: Chart Document that was created and inserted

        See Also:
            :py:class:`~.color.CommonColor`

        Hint:
            .. include:: ../../resources/utils/chart2_lookup_chart_tmpl.rst
        """
        try:
            # type check that diagram_name is ChartTemplateBase | str
            mInfo.Info.is_type_enum_multi(
                alt_type="str", enum_type=ChartTemplateBase, enum_val=diagram_name, arg_name="diagram_name"
            )
            chart_name = Chart2._CHART_NAME + str(int(random() * 10_000))
            cls.add_table_chart(
                sheet=sheet,
                chart_name=chart_name,
                cells_range=cells_range,
                cell_name=cell_name,
                width=width,
                height=height,
            )
            chart_doc = cls.get_chart_doc(sheet, chart_name)

            # assign chart template to the chart's diagram
            diagram = chart_doc.getFirstDiagram()
            ct_template = cls.set_template(chart_doc=chart_doc, diagram=diagram, diagram_name=diagram_name)

            has_cats = cls.has_categories(diagram_name)

            dp = chart_doc.getDataProvider()

            ps = mProps.Props.make_props(
                CellRangeRepresentation=mCalc.Calc.get_range_str(cells_range, sheet),
                DataRowSource=ChartDataRowSource.COLUMNS,
                FirstCellAsLabel=True,
                HasCategories=has_cats,
            )
            ds = dp.createDataSource(ps)

            # add data source to chart template
            args = mProps.Props.make_props(HasCategories=has_cats)
            ct_template.changeDiagramData(diagram, ds, args)

            # apply style settings to chart doc
            # background and wall colors
            cls.set_background_colors(chart_doc, color_bg, color_wall)

            if has_cats:
                cls.set_data_point_labels(chart_doc, DataPointLabelTypeKind.NUMBER)

            return chart_doc
        except mEx.ChartError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error inserting chart") from e

    @staticmethod
    def add_table_chart(
        sheet: XSpreadsheet, chart_name: str, cells_range: CellRangeAddress, cell_name: str, width: int, height: int
    ) -> None:
        """
        Adds a new table chart at a given cell name and size, using cells range.

        Args:
            sheet (XSpreadsheet): Spreadsheet
            chart_name (str): Chart name.
            cells_range (CellRangeAddress): Cell Range
            cell_name (str): Cell Name such as ``A1``
            width (int): Chart Width
            height (int): Chart Height

        Raises:
            ChartError: If error occurs.

        Returns:
            None:
        """
        try:
            charts_supp = mLo.Lo.qi(XTableChartsSupplier, sheet, True)
            tbl_charts = charts_supp.getCharts()

            pos = mCalc.Calc.get_cell_pos(sheet, cell_name)
            rect = Rectangle(X=pos.X, Y=pos.Y, Width=width * 1_000, Height=height * 1_000)
            addrs = (cells_range,)

            # 2nd last arg: whether the topmost row of the source data will
            # be used to set labels for the category axis or the legend;
            # last arg: whether the leftmost column of the source data will
            # be used to set labels for the category axis or the legend.
            tbl_charts.addNewByName(chart_name, rect, addrs, True, True)
        except Exception as e:
            raise mEx.ChartError("Error adding table chart") from e

    @staticmethod
    def set_template(
        chart_doc: XChartDocument, diagram: XDiagram, diagram_name: ChartTemplateBase | str
    ) -> XChartTypeTemplate:
        """
        Sets template of chart

        Args:
            chart_doc (XChartDocument): Chart Document
            diagram (XDiagram): diagram
            diagram_name (ChartTemplateBase | str): Diagram template name

        Raises:
            ChartError: If error occurs.

        Returns:
            XInterface: Chart Template

        Note:
            If unable to create template from ``diagram_name`` for any reason then
            ``ChartTypes.Column.TEMPLATE_STACKED.COLUMN`` (``"Column"``) is used as a fallback.

        Hint:
            .. include:: ../../resources/utils/chart2_lookup_chart_tmpl.rst
        """
        # XChartTypeTemplate does not seem to be supported by LO 7.4 ( gets import error )
        # Available interfaces com.sun.star.chart2.template.Column: (also XInterface)
        # com.sun.star.beans.XFastPropertySet
        # com.sun.star.beans.XMultiPropertySet
        # com.sun.star.beans.XMultiPropertyStates
        # com.sun.star.beans.XPropertySet
        # com.sun.star.beans.XPropertyState
        # com.sun.star.lang.XServiceName
        # com.sun.star.lang.XTypeProvider
        # com.sun.star.style.XStyleSupplier
        # com.sun.star.uno.XWeak

        # in LO 7.3 com.sun.star.chart2.XChartTypeTemplate is included

        # ensure diagram_name is ChartTemplateBase | str
        mInfo.Info.is_type_enum_multi(
            alt_type="str", enum_type=ChartTemplateBase, enum_val=diagram_name, arg_name="diagram_name"
        )

        try:
            ct_man = chart_doc.getChartTypeManager()
            msf = mLo.Lo.qi(XMultiServiceFactory, ct_man, True)
            template_nm = f"com.sun.star.chart2.template.{diagram_name}"
            ct_template = mLo.Lo.qi(XChartTypeTemplate, msf.createInstance(template_nm))
            if ct_template is None:
                mLo.Lo.print(f'Could not create chart template "{diagram_name}"; using a column chart instead')
                ct_template = mLo.Lo.qi(
                    XChartTypeTemplate, msf.createInstance("com.sun.star.chart2.template.Column"), True
                )

            ct_template.changeDiagram(diagram)
            return ct_template
        except Exception as e:
            raise mEx.ChartError("Error setting chart template") from e

    @staticmethod
    def has_categories(diagram_name: ChartTemplateBase | str) -> bool:
        """
        Gets if diagram name has categories

        Args:
            diagram_name (ChartTemplateBase | str): Diagram Name

        Returns:
            bool: ``True`` if has categories; Otherwise, ``False``.

        Hint:
            .. include:: ../../resources/utils/chart2_lookup_chart_tmpl.rst
        """
        # All the chart templates, except for scatter and bubble use
        # categories on the x-axis

        # Ensure diagram_name ChartTemplateBase | str
        mInfo.Info.is_type_enum_multi(
            alt_type="str", enum_type=ChartTemplateBase, enum_val=diagram_name, arg_name="diagram_name"
        )

        dn = str(diagram_name).lower()
        non_cats = ("scatter", "bubble")
        for non_cat in non_cats:
            if non_cat in dn:
                return False
        return True

    # endregion insert a chart

    # region get a chart
    @classmethod
    def get_chart_doc(cls, sheet: XSpreadsheet, chart_name: str) -> XChartDocument:
        """
        Gets the chart document from the sheet

        Args:
            sheet (XSpreadsheet): Spreadsheet
            chart_name (str): Chart Name

        Raises:
            ChartError: If error occurs

        Returns:
            XChartDocument: Spreadsheet document.
        """
        try:
            tbl_chart = cls.get_table_chart(sheet, chart_name)
            eos = mLo.Lo.qi(XEmbeddedObjectSupplier, tbl_chart, True)
            return mLo.Lo.qi(XChartDocument, eos.getEmbeddedObject(), True)
        except mEx.ChartError:
            raise
        except Exception as e:
            raise mEx.ChartError(f'Error getting chart document for chart "{chart_name}"') from e

    @staticmethod
    def get_table_chart(sheet: XSpreadsheet, chart_name: str) -> XTableChart:
        """
        Gets the named table chart from the sheet

        Args:
            sheet (XSpreadsheet): Spreadsheet
            chart_name (str): Chart Name

        Raises:
            ChartError: If error occurs.

        Returns:
            XTableChart: Table Chart
        """
        try:
            charts_supp = mLo.Lo.qi(XTableChartsSupplier, sheet, True)
            tbl_charts = charts_supp.getCharts()
            tc_access = mLo.Lo.qi(XNameAccess, tbl_charts, True)
            tbl_chart = mLo.Lo.qi(XTableChart, tc_access.getByName(chart_name))
            return tbl_chart
        except Exception as e:
            raise mEx.ChartError(f'Error getting table chart for chart "{chart_name}"') from e

    @staticmethod
    def get_chart_templates(chart_doc: XChartDocument) -> List[str]:
        """
        Gets a list of chart templates (services).

        Args:
            chart_doc (XChartDocument): Chart Document.

        Raises:
            ChartError: If error occurs

        Returns:
            List[str]: List of chart templates
        """
        try:
            ct_man = chart_doc.getChartTypeManager()
            return mInfo.Info.get_available_services(ct_man)
        except Exception as e:
            raise mEx.ChartError("Error getting chart templates") from e

    # endregion get a chart

    # region titles
    @classmethod
    def set_title(cls, chart_doc: XChartDocument, title: str) -> XTitle:
        """
        Sets the title of chart

        Args:
            chart_doc (XChartDocument): Chart Document.
            title (str): Title as string

        Raises:
            ChartError: If error occurs.

        Returns:
            XTitle: Title Object

        Note:
            The title is set to a font size of ``14`` and the font applied is
            the font returned by :py:meth:`.Info.get_font_general_name`

        Hint:
            The returned Title object can be passed to :py:meth:`~.Chart2.set_x_title_font` to
            change default font.
        """
        try:
            # return XTilte so it may have futher styles applied
            titled = mLo.Lo.qi(XTitled, chart_doc, True)
            xtitle = cls.create_title(title)
            titled.setTitleObject(xtitle)
            fname = mInfo.Info.get_font_general_name()
            cls.set_x_title_font(xtitle, fname, 14)
            return xtitle
        except Exception as e:
            raise mEx.ChartError("Error setting title for chart") from e

    @staticmethod
    def create_title(title: str) -> XTitle:
        """
        Creates a title object

        Args:
            title (str): Title text.

        Raises:
            ChartError: If error occurs.

        Returns:
            XTitle: Title object.
        """
        try:
            xtitle = mLo.Lo.create_instance_mcf(XTitle, "com.sun.star.chart2.Title", raise_err=True)
            xtitle_str = mLo.Lo.create_instance_mcf(
                XFormattedString, "com.sun.star.chart2.FormattedString", raise_err=True
            )
            xtitle_str.setString(title)
            title_arr = (xtitle_str,)
            xtitle.setText(title_arr)
            return xtitle
        except Exception as e:
            raise mEx.ChartError(f'Error creating title for: "{title}"') from e

    @staticmethod
    def set_x_title_font(xtitle: XTitle, font_name: str, pt_size: int) -> None:
        """
        Sets X title font.

        Args:
            xtitle (XTitle): Title instance.
            font_name (str): Font Name
            pt_size (int): Font point size

        Raises:
            ChartError: If error occurs.

        Returns:
            None:

        See Also:
            :py:meth:`.Info.get_font_general_name`
        """
        try:
            fo_strs = xtitle.getText()
            if fo_strs:
                mProps.Props.set_property(fo_strs[0], "CharFontName", font_name)
                mProps.Props.set_property(fo_strs[0], "CharHeight", pt_size)
        except Exception as e:
            raise mEx.ChartError("Error setting x title font") from e

    @staticmethod
    def get_title(chart_doc: XChartDocument) -> XTitle:
        """
        Gets Title from chart

        Args:
            chart_doc (XChartDocument): Chart Document.

        Raises:
            ChartError: If error occurs.

        Returns:
            XTitle: Title object.
        """
        try:
            xtilted = mLo.Lo.qi(XTitled, chart_doc, True)
            return xtilted.getTitleObject()
        except Exception as e:
            raise mEx.ChartError("Error getting title from chart") from e

    @classmethod
    def set_subtitle(cls, chart_doc: XChartDocument, subtitle: str) -> XTitle:
        """
        Gets subtitle

        Args:
            chart_doc (XChartDocument): Chart Document.
            subtitle (str): Subtitle text.

        Raises:
            ChartError: If error occurs

        Returns:
            XTitle: Title object

        Note:
            The subtitle is set to a font size of ``12`` and the font applied is
            the font returned by :py:meth:`.Info.get_font_general_name`

        Hint:
            The returned Title object can be passed to :py:meth:`~.Chart2.set_x_title_font` to
            change default font.
        """
        try:
            diagram = chart_doc.getFirstDiagram()
            titled = mLo.Lo.qi(XTitled, diagram, True)
            title = cls.create_title(subtitle)
            titled.setTitleObject(title)
            fname = mInfo.Info.get_font_general_name()
            cls.set_x_title_font(title, fname, 12)
            return title
        except mEx.ChartError:
            raise
        except Exception as e:
            raise mEx.ChartError(f'Error setting subtitle "{subtitle}" for chart') from e

    @staticmethod
    def get_subtitle(chart_doc: XChartDocument) -> XTitle:
        """
        Gets subtitle from chart

        Args:
            chart_doc (XChartDocument): Chart Document.

        Raises:
            ChartError: If error occurs.

        Returns:
            XTitle: Title object.
        """
        try:
            diagram = chart_doc.getFirstDiagram()
            titled = mLo.Lo.qi(XTitled, diagram, True)
            return titled.getTitleObject()
        except Exception as e:
            raise mEx.ChartError("Error getting subtitle from chart") from e

    # endregion titles

    # region Axis
    @classmethod
    def get_axis(cls, chart_doc: XChartDocument, axis_val: AxisKind, idx: int) -> XAxis:
        """
        Gets axis

        Args:
            chart_doc (XChartDocument): Chart Document
            axis_val (AxisKind): Axis Kind
            idx (int): Axis index

        Raises:
            ChartError: If error occurs.

        Returns:
            XAxis: Axis object.
        """
        try:
            coord_sys = cls.get_coord_system(chart_doc)
            result = coord_sys.getAxisByDimension(int(axis_val), idx)
            if result is None:
                raise mEx.UnKnownError("None Value: getAxisByDimension() returned None")
            return result
        except mEx.ChartError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error getting Axis for chart") from e

    @classmethod
    def get_x_axis(cls, chart_doc: XChartDocument) -> XAxis:
        """
        Get Chart X axis.

        Args:
            chart_doc (XChartDocument): Chart Document.

        Raises:
            ChartError: If error occurs.

        Returns:
            XAxis: Chart X Axis object.

        See Also:
            :py:meth:`~.Chart2.get_axis`
        """
        return cls.get_axis(chart_doc=chart_doc, axis_val=AxisKind.X, idx=0)

    @classmethod
    def get_y_axis(cls, chart_doc: XChartDocument) -> XAxis:
        """
        Get Chart Y axis.

        Args:
            chart_doc (XChartDocument): Chart Document.

        Raises:
            ChartError: If error occurs.

        Returns:
            XAxis: Chart Y Axis object.

        See Also:
            :py:meth:`~.Chart2.get_axis`
        """
        return cls.get_axis(chart_doc=chart_doc, axis_val=AxisKind.Y, idx=0)

    @classmethod
    def get_x_axis2(cls, chart_doc: XChartDocument) -> XAxis:
        """
        Get Chart X axis2.

        Args:
            chart_doc (XChartDocument): Chart Document.

        Raises:
            ChartError: If error occurs.

        Returns:
            XAxis: Chart X Axis2 object.

        See Also:
            :py:meth:`~.Chart2.get_axis`
        """
        return cls.get_axis(chart_doc=chart_doc, axis_val=AxisKind.X, idx=1)

    @classmethod
    def get_y_axis2(cls, chart_doc: XChartDocument) -> XAxis:
        """
        Get Chart Y axis2.

        Args:
            chart_doc (XChartDocument): Chart Document.

        Raises:
            ChartError: If error occurs.

        Returns:
            XAxis: Chart Y Axis2 object.

        See Also:
            :py:meth:`~.Chart2.get_axis`
        """
        return cls.get_axis(chart_doc=chart_doc, axis_val=AxisKind.Y, idx=1)

    @classmethod
    def set_axis_title(cls, chart_doc: XChartDocument, title: str, axis_val: AxisKind, idx: int) -> XTitle:
        """
        Sets axis title.

        Args:
            chart_doc (XChartDocument): Chart Document.
            title (str): Title text.
            axis_val (AxisKind): Axis kind.
            idx (int): Index

        Raises:
            ChartError: If error occurs.

        Returns:
            XTitle: Axis Title.

        Note:
            The title is set to a font size of ``12`` and the font applied is
            the font returned by :py:meth:`.Info.get_font_general_name`

        Hint:
            The returned Title object can be passed to :py:meth:`~.Chart2.set_x_title_font` to
            change default font.

        See Also:
            :py:meth:`~.Chart2.get_axis`
        """
        try:
            axis = cls.get_axis(chart_doc=chart_doc, axis_val=axis_val, idx=idx)
            titled_axis = mLo.Lo.qi(XTitled, axis, True)
            xtitle = cls.create_title(title)
            titled_axis.setTitleObject(xtitle)
            fname = mInfo.Info.get_font_general_name()
            cls.set_x_title_font(xtitle, fname, 12)
            return xtitle
        except mEx.ChartError:
            raise
        except Exception as e:
            raise mEx.ChartError(f'Error setting axis tile: "{title}" for chart') from e

    @classmethod
    def set_x_axis_title(cls, chart_doc: XChartDocument, title: str) -> XTitle:
        """
        Sets X axis Title

        Args:
            chart_doc (XChartDocument): Chart Document.
            title (str): Title Text

        Returns:
            XTitle: Title object

        See Also:
            :py:meth:`~.Chart2.set_axis_title`
        """
        return cls.set_axis_title(chart_doc=chart_doc, title=title, axis_val=AxisKind.X, idx=0)

    @classmethod
    def set_y_axis_title(cls, chart_doc: XChartDocument, title: str) -> XTitle:
        """
        Sets Y axis Title

        Args:
            chart_doc (XChartDocument): Chart Document.
            title (str): Title Text

        Returns:
            XTitle: Title object

        See Also:
            :py:meth:`~.Chart2.set_axis_title`
        """
        return cls.set_axis_title(chart_doc=chart_doc, title=title, axis_val=AxisKind.Y, idx=0)

    @classmethod
    def set_x_axis2_title(cls, chart_doc: XChartDocument, title: str) -> XTitle:
        """
        Sets X axis2 Title

        Args:
            chart_doc (XChartDocument): Chart Document.
            title (str): Title Text

        Returns:
            XTitle: Title object

        See Also:
            :py:meth:`~.Chart2.set_axis_title`
        """
        return cls.set_axis_title(chart_doc=chart_doc, title=title, axis_val=AxisKind.X, idx=1)

    @classmethod
    def set_y_axis2_title(cls, chart_doc: XChartDocument, title: str) -> XTitle:
        """
        Sets Y axis2 Title

        Args:
            chart_doc (XChartDocument): Chart Document.
            title (str): Title Text

        Returns:
            XTitle: Title object

        See Also:
            :py:meth:`~.Chart2.set_axis_title`
        """
        return cls.set_axis_title(chart_doc=chart_doc, title=title, axis_val=AxisKind.Y, idx=1)

    @classmethod
    def get_axis_title(cls, chart_doc: XChartDocument, axis_val: AxisKind, idx: int) -> XTitle:
        """
        Gets axis Title

        Args:
            chart_doc (XChartDocument): Chart Document
            axis_val (AxisKind): Axis Kind.
            idx (int): Index

        Raises:
            ChartError: If error occurs

        Returns:
            XTitle: Title object.

        See Also:
            :py:meth:`~.Chart2.get_axis`
        """
        try:
            axis = cls.get_axis(chart_doc=chart_doc, axis_val=axis_val, idx=idx)
            titled_axis = mLo.Lo.qi(XTitled, axis, True)
            result = titled_axis.getTitleObject()
            if result is None:
                raise mEx.UnKnownError("None Value: getTitleObject() return a value of None")
            return result
        except mEx.ChartError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error getting axis title") from e

    @classmethod
    def get_x_axis_title(cls, chart_doc: XChartDocument) -> XTitle:
        """
        Gets X axis title.

        Args:
            chart_doc (XChartDocument): Chart Document

        Raises:
            ChartError: If error occurs

        Returns:
            XTitle: Title object.

        See Also:
            :py:meth:`~.Chart2.get_axis_title`
        """
        return cls.get_axis_title(chart_doc=chart_doc, axis_val=AxisKind.X, idx=0)

    @classmethod
    def get_y_axis_title(cls, chart_doc: XChartDocument) -> XTitle:
        """
        Gets Y axis title.

        Args:
            chart_doc (XChartDocument): Chart Document

        Raises:
            ChartError: If error occurs

        Returns:
            XTitle: Title object.

        See Also:
            :py:meth:`~.Chart2.get_axis_title`
        """
        return cls.get_axis_title(chart_doc=chart_doc, axis_val=AxisKind.Y, idx=0)

    @classmethod
    def get_x_axis2_title(cls, chart_doc: XChartDocument) -> XTitle:
        """
        Gets X axis2 title.

        Args:
            chart_doc (XChartDocument): Chart Document

        Raises:
            ChartError: If error occurs

        Returns:
            XTitle: Title object.

        See Also:
            :py:meth:`~.Chart2.get_axis_title`
        """
        return cls.get_axis_title(chart_doc=chart_doc, axis_val=AxisKind.X, idx=1)

    @classmethod
    def get_y_axis2_title(cls, chart_doc: XChartDocument) -> XTitle:
        """
        Gets Y axis2 title.

        Args:
            chart_doc (XChartDocument): Chart Document

        Raises:
            ChartError: If error occurs

        Returns:
            XTitle: Title object.

        See Also:
            :py:meth:`~.Chart2.get_axis_title`
        """
        return cls.get_axis_title(chart_doc=chart_doc, axis_val=AxisKind.Y, idx=1)

    @classmethod
    def rotate_axis_title(cls, chart_doc: XChartDocument, axis_val: AxisKind, idx: int, angle: Angle) -> None:
        """
        Rotates axis title.

        Args:
            chart_doc (XChartDocument): Chart Document.
            axis_val (AxisKind): Axis kind.
            idx (int): Index
            angle (Angle): Angle

        Raises:
            ChartError: If error occurs.

        Returns:
            None:
        """
        try:
            xtitle = cls.get_axis_title(chart_doc=chart_doc, axis_val=axis_val, idx=idx)
            mProps.Props.set_property(xtitle, "TextRotation", angle.Value)
        except mEx.ChartError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error while trying to rotate axis title") from e

    @classmethod
    def rotate_x_axis_title(cls, chart_doc: XChartDocument, angle: Angle) -> None:
        """
        Rotates X axis title.

        Args:
            chart_doc (XChartDocument): Chart Document.
            angle (Angle): Angle

        Raises:
            ChartError: If error occurs.

        Returns:
            None:

        See Also:
            :py:meth:`~.Chart2.rotate_axis_title`
        """
        cls.rotate_axis_title(chart_doc=chart_doc, axis_val=AxisKind.X, idx=0, angle=angle)

    @classmethod
    def rotate_y_axis_title(cls, chart_doc: XChartDocument, angle: Angle) -> None:
        """
        Rotates Y axis title.

        Args:
            chart_doc (XChartDocument): Chart Document.
            angle (Angle): Angle

        Raises:
            ChartError: If error occurs.

        Returns:
            None:

        See Also:
            :py:meth:`~.Chart2.rotate_axis_title`
        """
        cls.rotate_axis_title(chart_doc=chart_doc, axis_val=AxisKind.Y, idx=0, angle=angle)

    @classmethod
    def rotate_x_axis2_title(cls, chart_doc: XChartDocument, angle: Angle) -> None:
        """
        Rotates X axis2 title.

        Args:
            chart_doc (XChartDocument): Chart Document.
            angle (Angle): Angle

        Raises:
            ChartError: If error occurs.

        Returns:
            None:

        See Also:
            :py:meth:`~.Chart2.rotate_axis_title`
        """
        cls.rotate_axis_title(chart_doc=chart_doc, axis_val=AxisKind.X, idx=1, angle=angle)

    @classmethod
    def rotate_y_axis2_title(cls, chart_doc: XChartDocument, angle: Angle) -> None:
        """
        Rotates Y axis2 title.

        Args:
            chart_doc (XChartDocument): Chart Document.
            angle (Angle): Angle

        Raises:
            ChartError: If error occurs.

        Returns:
            None:

        See Also:
            :py:meth:`~.Chart2.rotate_axis_title`
        """
        cls.rotate_axis_title(chart_doc=chart_doc, axis_val=AxisKind.Y, idx=1, angle=angle)

    @classmethod
    def show_axis_label(cls, chart_doc: XChartDocument, axis_val: AxisKind, idx: int, is_visible: bool) -> None:
        """
        Sets the visibility for chart axis label.

        Args:
            chart_doc (XChartDocument): Chart Document.
            axis_val (AxisKind): Axis kind.
            idx (int): Index
            is_visible (bool): Visible state

        Raises:
            ChartError: If error occurs.

        Returns:
            None:
        """
        try:
            axis = cls.get_axis(chart_doc=chart_doc, axis_val=axis_val, idx=idx)
            mProps.Props.set_property(axis, "Show", is_visible)
        except mEx.ChartError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error while setting axis label visibility") from e

    @classmethod
    def show_x_axis_label(cls, chart_doc: XChartDocument, is_visible: bool) -> None:
        """
        Sets the visibility for chart X axis label.

        Args:
            chart_doc (XChartDocument): Chart Document.
            is_visible (bool): Visible state

        Raises:
            ChartError: If error occurs.

        Returns:
            None:

        See Also:
            :py:meth:`~.Chart2.show_axis_label`
        """
        cls.show_axis_label(chart_doc=chart_doc, axis_val=AxisKind.X, idx=0, is_visible=is_visible)

    @classmethod
    def show_y_axis_label(cls, chart_doc: XChartDocument, is_visible: bool) -> None:
        """
        Sets the visibility for chart Y axis label.

        Args:
            chart_doc (XChartDocument): Chart Document.
            is_visible (bool): Visible state

        Raises:
            ChartError: If error occurs.

        Returns:
            None:

        See Also:
            :py:meth:`~.Chart2.show_axis_label`
        """
        cls.show_axis_label(chart_doc=chart_doc, axis_val=AxisKind.Y, idx=0, is_visible=is_visible)

    @classmethod
    def show_x_axis2_label(cls, chart_doc: XChartDocument, is_visible: bool) -> None:
        """
        Sets the visibility for chart X axis2 label.

        Args:
            chart_doc (XChartDocument): Chart Document.
            is_visible (bool): Visible state

        Raises:
            ChartError: If error occurs.

        Returns:
            None:

        See Also:
            :py:meth:`~.Chart2.show_axis_label`
        """
        cls.show_axis_label(chart_doc=chart_doc, axis_val=AxisKind.X, idx=1, is_visible=is_visible)

    @classmethod
    def show_y_axis2_label(cls, chart_doc: XChartDocument, is_visible: bool) -> None:
        """
        Sets the visibility for chart Y axis2 label.

        Args:
            chart_doc (XChartDocument): Chart Document.
            is_visible (bool): Visible state

        Raises:
            ChartError: If error occurs.

        Returns:
            None:

        See Also:
            :py:meth:`~.Chart2.show_axis_label`
        """
        cls.show_axis_label(chart_doc=chart_doc, axis_val=AxisKind.Y, idx=1, is_visible=is_visible)

    @classmethod
    def scale_axis(cls, chart_doc: XChartDocument, axis_val: AxisKind, idx: int, scale_type: CurveKind) -> XAxis:
        """
        Scales the chart axis.

        Args:
            chart_doc (XChartDocument): Chart Document.
            axis_val (AxisKind): Axis kind.
            idx (int): Index
            scale_type (CurveKind): Scale kind

        Raises:
            ChartError: If error occurs.

        Returns:
            XAxis: Axis object.

        Note:
            Supported types of ``scale_type`` are ``LINEAR``, ``LOGARITHMIC``, ``EXPONENTIAL`` and ``POWER``.
            If ``scale_type``  is not supported then the ``Scaling`` is not set.
        """
        try:
            axis = cls.get_axis(chart_doc=chart_doc, axis_val=axis_val, idx=idx)
            sd = axis.getScaleData()
            s = None
            if scale_type == CurveKind.LINEAR:
                s = "LinearScaling"
            elif scale_type == CurveKind.LOGARITHMIC:
                s = "LogarithmicScaling"
            elif scale_type == CurveKind.EXPONENTIAL:
                s = "ExponentialScaling"
            elif scale_type == CurveKind.POWER:
                s = "PowerScaling"
            if s is None:
                mLo.Lo.print(f'Did not reconize scaling type: "{scale_type}"')
            else:
                sd.Scaling = mLo.Lo.create_instance_mcf(XScaling, f"com.sun.star.chart2.{s}", raise_err=True)
            axis.setScaleData(sd)
            return axis
        except mEx.ChartError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error setting axis scale") from e

    @classmethod
    def scale_x_axis(cls, chart_doc: XChartDocument, scale_type: CurveKind) -> XAxis:
        """
        Scales the chart X axis.

        Args:
            chart_doc (XChartDocument): Chart Document.
            scale_type (CurveKind): Scale kind

        Raises:
            ChartError: If error occurs.

        Returns:
            XAxis: Axis object.

        Note:
            Supported types of ``scale_type`` are ``LINEAR``, ``LOGARITHMIC``, ``EXPONENTIAL`` and ``POWER``.
            If ``scale_type``  is not supported then the ``Scaling`` is not set.

        See Also:
            :py:meth:`~.Chart2.scale_axis`
        """
        return cls.scale_axis(chart_doc=chart_doc, axis_val=AxisKind.X, idx=0, scale_type=scale_type)

    @classmethod
    def scale_y_axis(cls, chart_doc: XChartDocument, scale_type: CurveKind) -> XAxis:
        """
        Scales the chart Y axis.

        Args:
            chart_doc (XChartDocument): Chart Document.
            scale_type (CurveKind): Scale kind

        Raises:
            ChartError: If error occurs.

        Returns:
            XAxis: Axis object.

        Note:
            Supported types of ``scale_type`` are ``LINEAR``, ``LOGARITHMIC``, ``EXPONENTIAL`` and ``POWER``.
            If ``scale_type``  is not supported then the ``Scaling`` is not set.

        See Also:
            :py:meth:`~.Chart2.scale_axis`
        """
        return cls.scale_axis(chart_doc=chart_doc, axis_val=AxisKind.Y, idx=0, scale_type=scale_type)

    @classmethod
    def print_scale_data(cls, axis_name: str, axis: XAxis) -> None:
        """
        Prints axis info to console

        Args:
            axis_name (str): Axis Name
            axis (XAxis): Axis
        """
        sd = axis.getScaleData()
        print(f"Scaled Data for: {axis_name}")
        print(f"  Minimum: {sd.Minimum}")
        print(f"  Maximum: {sd.Maximum}")
        print(f"  Origin: {sd.Origin}")

        if sd.Orientation == AxisOrientation.MATHEMATICAL:
            print("  Orientation: mathematical")
        else:
            print("  Orientation: reverse")
        print(f"  Scaling: {mInfo.Info.get_implementation_name(sd.Scaling)}")
        print(f"  AxisType: {cls.get_axis_type_string(AxisTypeKind(sd.AxisType))}")
        print(f"  AutoDateAxis: {sd.AutoDateAxis}")
        print(f"  ShiftedCategoryPosition: {sd.ShiftedCategoryPosition}")
        print(f"  IncrementData: {sd.IncrementData}")
        print(f"  TimeIncrement: {sd.TimeIncrement}")

    @staticmethod
    def get_axis_type_string(axis_type: AxisTypeKind) -> str:
        """
        Gets axis type as string.

        Args:
            axis_type (AxisTypeKind): Axis Type

        Raises:
            UnKnownError: If unable to determine ``axis_type``

        Returns:
            str: Axis type as string.
        """
        if axis_type == AxisTypeKind.REALNUMBER:
            return "real numbers"
        elif axis_type == AxisTypeKind.PERCENT:
            return "percentages"
        elif axis_type == AxisTypeKind.CATEGORY:
            return "categories"
        elif axis_type == AxisTypeKind.SERIES:
            return "series names"
        elif axis_type == AxisTypeKind.DATE:
            return "dates"
        else:
            raise mEx.UnKnownError("axis_type is of unknow type")

    # endregion Axis

    # region grid lines
    # region set_grid_lines()
    @overload
    @classmethod
    def set_grid_lines(cls, chart_doc: XChartDocument, axis_val: AxisKind) -> XPropertySet:
        ...

    @overload
    @classmethod
    def set_grid_lines(cls, chart_doc: XChartDocument, axis_val: AxisKind, idx: int) -> XPropertySet:
        ...

    @classmethod
    def set_grid_lines(cls, chart_doc: XChartDocument, axis_val: AxisKind, idx: int = 0) -> XPropertySet:
        """
        Set the grid lines for a chart.

        Args:
            chart_doc (XChartDocument): Chart Document.
            axis_val (AxisKind): Axis kind.
            idx (int, optional): Index. Defaults to 0.

        Raises:
            ChartError: If error occurs.

        Returns:
            XPropertySet: Property Set of Grid Properties.
        """
        try:
            axis = cls.get_axis(chart_doc=chart_doc, axis_val=axis_val, idx=idx)
            props = axis.getGridProperties()
            mProps.Props.set_property(props, "LineStyle", LineStyle.DASH)
            mProps.Props.set_property(props, "LineDashName", str(LineStyleNameKind.FINE_DOTTED))
            return props
        except mEx.ChartError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error setting gird lines for chart") from e

    # endregion set_grid_lines()

    # endregion grid lines

    # region legend
    @staticmethod
    def view_legend(chart_doc: XChartDocument, is_visible: bool) -> None:
        """
        Sets charts legend visibility.

        Args:
            chart_doc (XChartDocument): Chart Document.
            is_visible (bool): Visible State

        Raises:
            ChartError: If error occurs.

        Returns:
            None:
        """
        try:
            diagram = chart_doc.getFirstDiagram()
            legend = diagram.getLegend()
            if is_visible and legend is None:
                leg = mLo.Lo.create_instance_mcf(XLegend, "com.sun.star.chart2.Legend", raise_err=True)
                mProps.Props.set_property(leg, "LineStyle", LineStyle.NONE)
                mProps.Props.set_property(leg, "FillStyle", FillStyle.SOLID)
                mProps.Props.set_property(leg, "FillTransparence", 100)
                diagram.setLegend(leg)

            mProps.Props.set_property(leg, "Show", is_visible)
        except Exception as e:
            raise mEx.ChartError("Error while setting legend visibility") from e

    # endregion legend

    # region background colors
    @staticmethod
    def set_background_colors(chart_doc: XChartDocument, bg_color: mColor.Color, wall_color: mColor.Color) -> None:
        """
        Set the background colors for a chart

        Args:
            chart_doc (XChartDocument): Chart Document.
            bg_color (Color): Color Value for background
            wall_color (Color): Color Value for wall

        Raises:
            ChartError: If error occurs.

        Returns:
            None:

        See Also:
            :py:class:`~.color.CommonColor`
        """
        try:
            if int(bg_color) > 0:
                bg_ps = chart_doc.getPageBackground()
                # bg_dpp = cast("DataPointProperties", bg_ps)
                # bg_dpp.FillBackground = True
                # bg_dpp.FillStyle = FillStyle.SOLID
                # bg_dpp.FillColor = int(bg_color)
                #
                # there is a bug with chart_doc.getPageBackground()
                # it is suppose to return XProperySet, which it does but
                # XProperySet methods such as  setPropertyValue and getPropertySetInfo are missing.
                # see also: Props._set_by_attribute()
                #
                # mProps.Props.show_props("Background", bg_ps)
                mProps.Props.set(bg_ps, FillBackground=True, FillStyle=FillStyle.SOLID, FillColor=int(bg_color))

            if int(wall_color) > 0:
                diagram = chart_doc.getFirstDiagram()
                wall_ps = diagram.getWall()
                # wall_dpp = cast("DataPointProperties", wall_ps)
                # wall_dpp.FillBackground = True
                # wall_dpp.FillStyle = FillStyle.SOLID
                # wall_dpp.FillColor = int(wall_color)
                #
                # there is a bug with diagram.getWall()
                # it is suppose to return XProperySet, which it does but
                # XProperySet methods such as  setPropertyValue and getPropertySetInfo are missing.
                # see also: Props._set_by_attribute()
                #
                # mProps.Props.show_props("Wall", wall_ps)
                mProps.Props.set(wall_ps, FillBackground=True, FillStyle=FillStyle.SOLID, FillColor=int(wall_color))
        except Exception as e:
            raise mEx.ChartError("Error setting background colors") from e

    # endregion background colors

    # region access data source and series
    @staticmethod
    def create_data_series() -> XDataSeries:
        """
        Gets an instance of Chart2 DataSeries.

        Raises:
            ChartError: If error occurs.

        Returns:
            XDataSeries: Data Series instance.
        """
        try:
            ds = mLo.Lo.create_instance_mcf(XDataSeries, "com.sun.star.chart2.DataSeries", raise_err=True)
            return ds
        except Exception as e:
            raise mEx.ChartError("Error, unable to create XDataSeries interface") from e

    # region get_data_series()
    @overload
    @classmethod
    def get_data_series(cls, chart_doc: XChartDocument) -> Tuple[XDataSeries, ...]:
        ...

    @overload
    @classmethod
    def get_data_series(cls, chart_doc: XChartDocument, chart_type: ChartTypeNameBase) -> Tuple[XDataSeries, ...]:
        ...

    @overload
    @classmethod
    def get_data_series(cls, chart_doc: XChartDocument, chart_type: str) -> Tuple[XDataSeries, ...]:
        ...

    @classmethod
    def get_data_series(
        cls, chart_doc: XChartDocument, chart_type: ChartTypeNameBase | str = ""
    ) -> Tuple[XDataSeries, ...]:
        """
        Gets data series for a chart of a given chart type.

        Args:
            chart_doc (XChartDocument): Chart Document
            chart_type (ChartTypeNameBase | str, optional): Chart Type.

        Raises:
            NotFoundError: If chart is not found
            ChartError: If any other error occurs.

        Returns:
            Tuple[XDataSeries, ...]: Data Series

        Hint:
            .. include:: ../../resources/utils/chart2_lookup_chart_name.rst

        See Also:
            :py:meth:`.Chart2.find_chart_type`
        """
        try:
            if chart_type:
                xchart_type = cls.find_chart_type(chart_doc, chart_type)
            else:
                xchart_type = cls.get_chart_type(chart_doc)
            ds_con = mLo.Lo.qi(XDataSeriesContainer, xchart_type, True)
            return ds_con.getDataSeries()
        except Exception as e:
            raise mEx.ChartError("Error getting chart data series") from e

    # endregion get_data_series()

    # region get_data_source()
    @overload
    @classmethod
    def get_data_source(cls, chart_doc: XChartDocument) -> XDataSource:
        ...

    @overload
    @classmethod
    def get_data_source(cls, chart_doc: XChartDocument, chart_type: ChartTypeNameBase) -> XDataSource:
        ...

    @overload
    @classmethod
    def get_data_source(cls, chart_doc: XChartDocument, chart_type: str) -> XDataSource:
        ...

    @classmethod
    def get_data_source(cls, chart_doc: XChartDocument, chart_type: ChartTypeNameBase | str = "") -> XDataSource:
        """
        Get data source of a chart for a given chart type.

        This method assumes that the programmer wants the first data source in the data series.
        This is adequate for most charts which only use one data source.

        Args:
            chart_doc (XChartDocument): Chart Document
            chart_type (ChartTypeNameBase | str): Chart type.

        Raises:
            NotFoundError: If chart is not found
            ChartError: If any other error occurs.

        Returns:
            XDataSource: Chart data source

        Hint:
            .. include:: ../../resources/utils/chart2_lookup_chart_name.rst

        See Also:
            :py:meth:`~.Chart2.get_data_series`
        """
        try:
            dsa = cls.get_data_series(chart_doc=chart_doc, chart_type=chart_type)
            ds = mLo.Lo.qi(XDataSource, dsa[0], True)
            return ds
        except mEx.NotFoundError:
            raise
        except mEx.ChartError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error getting data source for chart") from e

    # endregion get_data_source()

    # endregion access data source and series

    # region chart types
    @staticmethod
    def get_coord_system(chart_doc: XChartDocument) -> XCoordinateSystem:
        """
        Gets coordinate system

        Args:
            chart_doc (XChartDocument): Chart Document.

        Raises:
            ChartError: If error occurs.

        Returns:
            XCoordinateSystem: Coordinate system object.
        """
        try:
            diagram = chart_doc.getFirstDiagram()
            coord_sys_con = mLo.Lo.qi(XCoordinateSystemContainer, diagram, True)
            coord_sys = coord_sys_con.getCoordinateSystems()
            if coord_sys:
                if len(coord_sys) > 1:
                    mLo.Lo.print(f"No. of coord systems: {len(coord_sys)}; using first.")
            return coord_sys[0]  # will raise error if coord_sys is empyt or none
        except Exception as e:
            raise mEx.ChartError("Error unable to get coord_system") from e

    @classmethod
    def get_chart_types(cls, chart_doc: XChartDocument) -> Tuple[XChartType, ...]:
        """
        Gets chart types for a chart.

        Args:
            chart_doc (XChartDocument): Chart Document.

        Raises:
            ChartError: If error occurs.

        Returns:
            Tuple[XChartType, ...]: Tuple of chart types.

        See Also:
            :py:meth:`~.Chart2.get_chart_type`
        """
        try:
            coord_sys = cls.get_coord_system(chart_doc)
            ct_con = mLo.Lo.qi(XChartTypeContainer, coord_sys, True)
            result = ct_con.getChartTypes()
            if result is None:
                raise mEx.UnKnownError("None Value: getChartTypes() returned a value of None")
            return result
        except mEx.ChartError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error getting chart types") from e

    @classmethod
    def get_chart_type(cls, chart_doc: XChartDocument) -> XChartType:
        """
        Gets a chart type for a chart

        Args:
            chart_doc (XChartDocument): Chart Document.

        Raises:
            ChartError: If error occurs.

        Returns:
            XChartType: Chart type object.

        See Also:
            :py:meth:`~.Chart2.get_chart_types`
        """
        try:
            chart_types = cls.get_chart_types(chart_doc)
            return chart_types[0]
        except mEx.ChartError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error getting chart type") from e

    @classmethod
    def find_chart_type(cls, chart_doc: XChartDocument, chart_type: ChartTypeNameBase | str) -> XChartType:
        """
        Finds a chart for a given chart type.

        Args:
            chart_doc (XChartDocument): Chart Document
            chart_type (ChartTypeNameBase | str): Chart type.

        Raises:
            NotFoundError: If chart is not found
            ChartError: If any other error occurs.

        Returns:
            XChartType: Found chart type.

        Hint:
            .. include:: ../../resources/utils/chart2_lookup_chart_name.rst
        """
        # Ensure chart_type is ChartTypeNameBase | str
        mInfo.Info.is_type_enum_multi(
            alt_type="str", enum_type=ChartTypeNameBase, enum_val=chart_type, arg_name="chart_type"
        )
        try:
            srch_name = f"com.sun.star.chart2.{str(chart_type).lower()}"
            chart_types = cls.get_chart_types(chart_doc)
            for ct in chart_types:
                ct_name = ct.getChartType().lower()
                if ct_name == srch_name:
                    return ct
        except Exception as e:
            raise mEx.ChartError(f'Error Finding chart for "{chart_type}"') from e
        raise mEx.NotFoundError(f'Chart for type "{chart_type}" was not found')

    @classmethod
    def print_chart_types(cls, chart_doc: XChartDocument) -> None:
        """
        Prints chart types to the console

        Args:
            chart_doc (XChartDocument): Chart Document.
        """
        chart_types = cls.get_chart_types(chart_doc)
        if len(chart_types) > 1:
            print(f"No. of chart tyeps: {len(chart_types)}")
            for ct in chart_types:
                print(f"  {ct.getChartType()}")
        else:
            print(f"Chart Type: {chart_types[0].getChartType()}")
        print()

    # region add_chart_type()
    @overload
    @classmethod
    def add_chart_type(cls, chart_doc: XChartDocument, chart_type: ChartTypeNameBase) -> XChartType:
        ...

    @overload
    @classmethod
    def add_chart_type(cls, chart_doc: XChartDocument, chart_type: str) -> XChartType:
        ...

    @classmethod
    def add_chart_type(cls, chart_doc: XChartDocument, chart_type: ChartTypeNameBase | str) -> XChartType:
        """
        Adds a chart type.

        Args:
            chart_doc (XChartDocument): Chart Document
            chart_type (ChartTypeNameBase | str): Chart type.

        Raises:
            ChartError: If error occurs

        Returns:
            XChartType: Chart type object

        Hint:
            .. include:: ../../resources/utils/chart2_lookup_chart_name.rst
        """
        mInfo.Info.is_type_enum_multi(
            alt_type="str", enum_type=ChartTypeNameBase, enum_val=chart_type, arg_name="chart_type"
        )
        try:
            ct = mLo.Lo.create_instance_mcf(XChartType, f"com.sun.star.chart2.{chart_type}", raise_err=True)
            coord_sys = cls.get_coord_system(chart_doc)
            ct_con = mLo.Lo.qi(XChartTypeContainer, coord_sys, True)
            ct_con.addChartType(ct)
            return ct
        except mEx.ChartError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error adding chart type") from e

    # endregion add_chart_type()

    # endregion chart types

    # region using a data source
    @staticmethod
    def show_data_source_args(chart_doc: XChartDocument, data_source: XDataSource) -> None:
        """
        Prints data information to console

        Args:
            chart_doc (XChartDocument): Chart Document.
            data_source (XDataSource): Data Source.

        Returns:
            None:
        """
        dp = chart_doc.getDataProvider()
        ps = dp.detectArguments(data_source)
        mProps.Props.show_props("Data Source arguments", ps)

    @staticmethod
    def print_labeled_seqs(data_source: XDataSource) -> None:
        """
        Prints labeled sequence information to console.

        A diagnostic function for printing all the labeled data sequences stored in an XDataSource:

        Args:
            data_source (XDataSource): Data Source.

        Returns:
            None:
        """
        data_seqs = data_source.getDataSequences()
        print(f"No. of sequeneces in data source: {len(data_seqs)}")
        for seq in data_seqs:
            label_seq = seq.getLabel().getData()
            print(f"{label_seq[0]} :")
            vals_seq = seq.getValues().getData()
            for val in vals_seq:
                print(f"  {val}")
            print()
            sr_rep = seq.getValues().getSourceRangeRepresentation()
            print(f"  Source range: {sr_rep}")
        print()

    @staticmethod
    def get_chart_data(data_source: XDataSource, idx: int) -> Tuple[float]:
        """
        Gets chart data

        Args:
            data_source (XDataSource): Data Source.
            idx (int): Index

        Raises:
            IndexError: If index is out of range.
            ChartError: If any other error occurs.

        Returns:
            Tuple[float]: Chart data.
        """
        try:
            data_seqs = data_source.getDataSequences()
            if idx < 0 or idx >= len(data_seqs):
                raise IndexError(f"Index value of {idx} is out of of range")

            vals_seq = data_seqs[idx].getValues().getData()
            vals: List[float] = []
            for val in vals_seq:
                vals.append(float(val))
            return tuple(vals)
        except IndexError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error getting chart data") from e

    # endregion using a data source

    # region using data series point props
    @classmethod
    def get_data_points_props(cls, chart_doc: XChartDocument, idx: int) -> List[XPropertySet]:
        """
        Gets all the properties for the data in the specified series

        Args:
            chart_doc (XChartDocument): Chart Document.
            idx (int): Index

        Raises:
            IndexError: If idx is out of range.

        Returns:
            List[XPropertySet]: Property set list.
        """
        data_series_arr = cls.get_data_series(chart_doc=chart_doc)
        if idx < 0 or idx >= len(data_series_arr):
            raise IndexError(f"Index value of {idx} is out of of range")

        props_lst: List[XPropertySet] = []
        i = 0
        while True:
            try:
                props = data_series_arr[idx].getDataPointByIndex(i)
                if props is not None:
                    props_lst.append(props)
                i += 1
            except Exception:
                props = None

            if props is None:
                break
        if len(props_lst) > 0:
            mLo.Lo.print(f"No Series at index {idx}")
        return props_lst

    @classmethod
    def get_data_point_props(cls, chart_doc: XChartDocument, series_idx: int, idx: int) -> XPropertySet:
        """
        Get the proprieties for a specific index within the data points.

        Args:
            chart_doc (XChartDocument): Chart Document.
            series_idx (int): Series Index
            idx (int): Index to extract from the datapoints data

        Raises:
            NotFoundError: If ``series_idx`` did not find any data points.
            IndexError: If an index is out of range.

        Returns:
            XPropertySet: A single property set from the data points series.

        See Also:
            :py:meth:`~.Chart2.get_data_points_props`
        """
        props = cls.get_data_points_props(chart_doc=chart_doc, idx=series_idx)
        if not props:
            raise mEx.NotFoundError("No Datapoints found to get XPropertySet from")

        if idx < 0 or idx >= len(props):
            raise IndexError(f"Index value of {idx} is out of of range")

        return props[idx]

    @classmethod
    def set_data_point_labels(cls, chart_doc: XChartDocument, label_type: DataPointLabelTypeKind) -> None:
        """
        Sets the data point label of a chart

        Args:
            chart_doc (XChartDocument): Chart Document
            label_type (DataPointLabelTypeKind): Label Type

        Raises:
            ChartError: If error occurs.

        Returns:
            None:
        """
        try:
            data_series_arr = cls.get_data_series(chart_doc=chart_doc)
            for data_series in data_series_arr:
                dp_label = cast(DataPointLabel, mProps.Props.get_property(data_series, "Label"))
                dp_label.ShowNumber = False
                dp_label.ShowCategoryName = False
                dp_label.ShowLegendSymbol = False
                if label_type == DataPointLabelTypeKind.NUMBER:
                    dp_label.ShowNumber = True
                elif label_type == DataPointLabelTypeKind.PERCENT:
                    dp_label.ShowNumber = True
                    dp_label.ShowNumberInPercent = True
                elif label_type == DataPointLabelTypeKind.CATEGORY:
                    dp_label.ShowCategoryName = True
                elif label_type == DataPointLabelTypeKind.SYMBOL:
                    dp_label.ShowLegendSymbol = True
                elif label_type == DataPointLabelTypeKind.NONE:
                    pass
                else:
                    raise mEx.UnKnownError("label_type is of unknow type")

                mProps.Props.set_property(data_series, "Label", dp_label)
        except mEx.ChartError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error setting data point labels") from e

    @classmethod
    def set_chart_shape_3d(cls, chart_doc: XChartDocument, shape: DataPointGeometry3DEnum) -> None:
        """
        Sets chart 3d shape

        Args:
            chart_doc (XChartDocument): Chart Document.
            shape (DataPointGeometry3DEnum): Shape kind

        Raises:
            ChartError: If an error occurs.

        Returns:
            None:
        """
        try:
            data_series_arr = cls.get_data_series(chart_doc=chart_doc)
            for data_series in data_series_arr:
                mProps.Props.set_property(data_series, "Geometry3D", int(shape))
        except mEx.ChartError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error setting chart shape 3d") from e

    @classmethod
    def dash_lines(cls, chart_doc: XChartDocument) -> None:
        """
        Sets chart data series to dashed lines.

        Args:
            chart_doc (XChartDocument): Chart Document

        Raises:
            ChartError: If error occurs.

        Returns:
            None:
        """
        try:
            data_series_arr = cls.get_data_series(chart_doc=chart_doc)
            for data_series in data_series_arr:
                mProps.Props.set_property(data_series, "LineStyle", LineStyle.DASH)
                mProps.Props.set_property(data_series, "LineDashName", str(LineStyleNameKind.FINE_DASHED))
        except mEx.ChartError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error setting chart sash lines") from e

    @staticmethod
    def color_stock_bars(ct: XChartType, w_day_color: mColor.Color, b_day_color: mColor.Color) -> None:
        """
        Set color of stock bars for a ``CandleStickChartType`` chart.

        Args:
            ct (XChartType): Chart Type.
            w_day_color (Color): Chart white day color
            b_day_color (Color): Chart black day color

        Raises:
            NotSupportedError: If Chart is not of type ``CandleStickChartType``
            ChartError: If any other error occurs.

        Returns:
            None:

        See Also:
            :py:class:`~.color.CommonColor`
        """
        try:
            if ct.getChartType() == "com.sun.star.chart2.CandleStickChartType":
                # there is a bug with white_day_ps and black_day_ps
                # they are suppose br XProperySet, which they are but
                # XProperySet methods such as  setPropertyValue and getPropertySetInfo are missing.
                # see also: Props._set_by_attribute()
                white_day_ps = mLo.Lo.qi(XPropertySet, mProps.Props.get(ct, "WhiteDay"), True)
                white_day_dpp = cast("DataPointProperties", white_day_ps)
                white_day_dpp.FillColor = int(w_day_color)
                # mProps.Props.set_property(white_day_ps, "FillColor", int(w_day_color))

                black_day_ps = mLo.Lo.qi(XPropertySet, mProps.Props.get(ct, "BlackDay"), True)
                black_day_dpp = cast("DataPointProperties", black_day_ps)
                black_day_dpp.FillColor = int(b_day_color)
                # mProps.Props.set_property(black_day_ps, "FillColor", int(b_day_color))
            else:
                raise mEx.NotSupportedError(
                    f'Only candel stick charts supported. "{ct.getChartType()}" not supported.'
                )
        except mEx.NotSupportedError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error coloring stock bars") from e

    # endregion using data series point props

    # region regression
    @staticmethod
    def create_curve(curve_kind: CurveKind) -> XRegressionCurve:
        """
        Creates a regression curve.

        Matches the regression constants defined in ``curve_kind`` to regression services offered by the API:

        Args:
            curve_kind (CurveKind): Curve kind.

        Raises:
            ChartError: If error occurs.

        Returns:
            XRegressionCurve: Regression Curve object.
        """
        try:
            rc = mLo.Lo.create_instance_mcf(XRegressionCurve, curve_kind.to_namespace(), raise_err=True)
            return rc
        except Exception as e:
            raise mEx.ChartError("Error creating curve") from e

    @classmethod
    def draw_regression_curve(cls, chart_doc: XChartDocument, curve_kind: CurveKind) -> None:
        """
        Draws a regression curve.

        Args:
            chart_doc (XChartDocument): Chart Document
            curve_kind (CurveKind): Curve kind.

        Raises:
            ChartError: If error occurs.

        Returns:
            None:
        """
        try:
            data_series_arr = cls.get_data_series(chart_doc=chart_doc)
            rc_con = mLo.Lo.qi(XRegressionCurveContainer, data_series_arr[0], True)
            curve = cls.create_curve(curve_kind)
            rc_con.addRegressionCurve(curve)

            ps = curve.getEquationProperties()
            mProps.Props.set_property(ps, "ShowCorrelationCoefficient", True)
            mProps.Props.set_property(ps, "ShowEquation", True)

            key = cls.get_number_format_key(chart_doc=chart_doc, nf_str="0.00")  # 2 dp
            if key != -1:
                mProps.Props.set_property(ps, "NumberFormat", key)
        except mEx.ChartError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error drawing regression curve") from e

    @staticmethod
    def get_number_format_key(chart_doc: XChartDocument, nf_str: str) -> int:
        """
        Converts a number format string into a number format key, which can be assigned to
        ``NumberFormat`` property.

        Args:
            chart_doc (XChartDocument): Chart Document
            nf_str (str): Number format string.

        Raises:
            ChartError: If error occurs.

        Returns:
            int: Number format key.

        Note:
            The string-to-key conversion is straight forward if you know what number format string to use,
            but there's little documentation on them. Probably the best approach is to use the Format
            Cells menu item in a spreadsheet document, and examine the dialog
        """
        try:
            xfs = mLo.Lo.qi(XNumberFormatsSupplier, chart_doc, True)
            n_formats = xfs.getNumberFormats()
            key = int(n_formats.queryKey(nf_str, Locale("en", "us", ""), False))
            if key == -1:
                mLo.Lo.print(f'Could not access key for number format: "{nf_str}"')
            return key
        except Exception as e:
            raise mEx.ChartError("Error getting number format key") from e

    @staticmethod
    def get_curve_type(curve: XRegressionCurve) -> CurveKind:
        """
        Gets curve kind from regression object.

        Args:
            curve (XRegressionCurve): Regression curve object.

        Raises:
            NotFoundError: If unable to detect curve kind.

        Returns:
            CurveKind: Curve Kind
        """
        services = set(mInfo.Info.get_services(curve))
        if CurveKind.LINEAR.to_namespace() in services:
            return CurveKind.LINEAR
        elif CurveKind.LOGARITHMIC.to_namespace() in services:
            return CurveKind.LOGARITHMIC
        elif CurveKind.EXPONENTIAL.to_namespace() in services:
            return CurveKind.EXPONENTIAL
        elif CurveKind.POWER.to_namespace() in services:
            return CurveKind.POWER
        elif CurveKind.POLYNOMIAL.to_namespace() in services:
            return CurveKind.POLYNOMIAL
        elif CurveKind.MOVING_AVERAGE.to_namespace() in services:
            return CurveKind.MOVING_AVERAGE
        else:
            raise mEx.NotFoundError("Could not identify trend type of curve")

    @classmethod
    def eval_curve(cls, chart_doc: XChartDocument, curve: XRegressionCurve) -> None:
        """
        Uses ``XRegressionCurve.getCalculator()`` to access the ``XRegressionCurveCalculator`` interface.
        It sets up the data and parameters for a particular curve, and prints the results of curve fitting to the console.

        Args:
            chart_doc (XChartDocument): Chart Document
            curve (XRegressionCurve): Regression Curve object.
        """
        curve_calc = curve.getCalculator()
        degree = 1
        ct = cls.get_curve_type(curve)
        if ct != CurveKind.LINEAR:
            degree = 2  # assumes POLYNOMIAL trend has degree == 2

        # degree, forceIntercept, interceptValue, period (for moving average)
        # the last are for setRegressionProperties is movingType
        #   See: https://api.libreoffice.org/docs/idl/ref/namespacecom_1_1sun_1_1star_1_1chart2_1_1MovingAverageType.html
        # movingType Only if regression type is "Moving Average" 1, 3 or 3
        #   see: https://api.libreoffice.org/docs/idl/ref/interfacecom_1_1sun_1_1star_1_1chart2_1_1XRegressionCurveCalculator.html#ae65112de1214e140d9ce7b28ffb09292
        # Because thi sis not a Moving Average setting to 0
        curve_calc.setRegressionProperties(degree, False, 0.0, 2, 0)

        data_source = cls.get_data_source(chart_doc)
        # cls.print_labled_seqs(data_source)

        xvals = cls.get_chart_data(data_source=data_source, idx=0)
        yvals = cls.get_chart_data(data_source=data_source, idx=0)
        curve_calc.recalculateRegression(xvals, yvals)

        print(f"  Curve equations: {curve_calc.getRepresentation()}")
        cc = curve_calc.getCorrelationCoefficient()
        print(f"  R^2 value: {(cc*cc):.3f}")

    @classmethod
    def calc_regressions(cls, chart_doc: XChartDocument) -> None:
        """
        Several different regression functions are calculated using the chart's data.
        Their equations and ``R2`` values are printed to the console

        Args:
            chart_doc (XChartDocument): Chart Document
        """

        def curve_info(curve_kind: CurveKind) -> None:
            curve = cls.create_curve(curve_kind=curve_kind)
            print(f"{curve_kind.label} regression curve:")
            cls.eval_curve(chart_doc=chart_doc, curve=curve)
            print()

        curve_info(CurveKind.LINEAR)
        curve_info(CurveKind.LOGARITHMIC)
        curve_info(CurveKind.EXPONENTIAL)
        curve_info(CurveKind.POWER)
        curve_info(CurveKind.POLYNOMIAL)
        curve_info(CurveKind.MOVING_AVERAGE)

    # endregion regression

    # region add data to a chart
    @staticmethod
    def create_ld_seq(
        dp: XDataProvider, role: DataRoleKind | str, data_label: str, data_range: str
    ) -> XLabeledDataSequence:
        """
        Creates a ``XLabeledDataSequence`` instance from two ``XDataSequence`` objects,
        one acting as a label the other as data.

        The ``XDataSequence`` object representing the data must have its "Role" property
        set to indicate the type of the data.

        Args:
            dp (XDataProvider): Data Provider
            role (DataRoleKind | str): Role that indicate the type of the data.
            data_label (str): Data label
            data_range (str): Data range

        Raises:
            ChartError: If error occurs

        Returns:
            XLabeledDataSequence: Labeled data sequence object.
        """
        # reate labeled data sequence using label and data;
        # the data is for the specified role

        try:
            # create data sequence for the label
            lbl_seq = dp.createDataSequenceByRangeRepresentation(data_label)

            # reate data sequence for the data and role
            data_seq = dp.createDataSequenceByRangeRepresentation(data_range)

            ds_ps = mLo.Lo.qi(XPropertySet, data_seq, True)

            # specify data role (type)
            mProps.Props.set_property(ds_ps, "Role", str(role))
            # mProps.Props.show_props("Data Sequence", ds_ps)

            # create new labeled data sequence using sequences
            ld_seq = mLo.Lo.create_instance_mcf(
                XLabeledDataSequence, "com.sun.star.chart2.data.LabeledDataSequence", raise_err=True
            )
            ld_seq.setLabel(lbl_seq)
            ld_seq.setValues(data_seq)
            return ld_seq
        except Exception as e:
            raise mEx.ChartError("Error creating LD sequence") from e

    @classmethod
    def set_y_error_bars(cls, chart_doc: XChartDocument, data_label: str, data_range: str) -> None:
        """
        Sets Y error Bars

        Args:
            chart_doc (XChartDocument): Chart Document
            data_label (str): Data Label
            data_range (str): Data Range

        Raises:
            ChartError: If error occurs.

        Returns:
            None:
        """
        try:
            error_bars_ps = mLo.Lo.create_instance_mcf(XPropertySet, "com.sun.star.chart2.ErrorBar", raise_err=True)
            mProps.Props.set(
                error_bars_ps, ShowPositiveError=True, ShowNegativeError=True, ErrorBarStyle=ErrorBarStyle.FROM_DATA
            )

            # convert into data sink
            data_sink = mLo.Lo.qi(XDataSink, error_bars_ps, True)

            # use data provider to create labelled data sequences
            # for the +/- error ranges
            dp = chart_doc.getDataProvider()

            pos_err_seq = cls.create_ld_seq(
                dp=dp, role=DataRoleKind.ERROR_BARS_Y_POSITIVE, data_label=data_label, data_range=data_range
            )
            neg_err_seq = cls.create_ld_seq(
                dp=dp, role=DataRoleKind.ERROR_BARS_Y_NEGATIVE, data_label=data_label, data_range=data_range
            )

            ld_seq = (pos_err_seq, neg_err_seq)

            # store the error bar data sequences in the data sink
            data_sink.setData(ld_seq)
            # mProps.Props.show_obj_props("Error Bar", error_bars_ps)
            # "ErrorBarRangePositive" and "ErrorBarRangeNegative"
            # will now have ranges they are read-only

            # store error bar in data series
            data_series_arr = cls.get_data_series(chart_doc=chart_doc)
            # print(f'No. of data serice: {len(data_series_arr)}')
            data_series = data_series_arr[0]
            # mProps.Props.show_obj_props("Data Series 0", data_series)
            mProps.Props.set(data_series, ErrorBarY=error_bars_ps)
        except mEx.ChartError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error Setting y error bars") from e

    @classmethod
    def add_stock_line(cls, chart_doc: XChartDocument, data_label: str, data_range: str) -> None:
        """
        Add stock line to chart.

        Args:
            chart_doc (XChartDocument): Chart Document
            data_label (str): Data label
            data_range (str): Data range

        Raises:
            ChartError: If error occurs.

        Returns:
            None:
        """
        try:
            # add (empty) line chart to the doc
            ct = cls.add_chart_type(chart_doc=chart_doc, chart_type=ChartTypes.Line.NAMED.LINE_CHART)
            data_series_cnt = mLo.Lo.qi(XDataSeriesContainer, ct, True)

            # create (empty) data series in the line chart
            ds = mLo.Lo.create_instance_mcf(XDataSeries, "com.sun.star.chart2.DataSeries", raise_err=True)

            mProps.Props.set_property(ds, "Color", int(mColor.CommonColor.RED))
            data_series_cnt.addDataSeries(ds)

            # add data to series by treating it as a data sink
            data_sink = mLo.Lo.qi(XDataSink, ds, True)

            # add data as y values
            dp = chart_doc.getDataProvider()
            dl_seq = cls.create_ld_seq(dp=dp, role=DataRoleKind.VALUES_Y, data_label=data_label, data_range=data_range)
            ld_seq_arr = (dl_seq,)
            data_sink.setData(ld_seq_arr)
        except mEx.ChartError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error adding stock line") from e

    @classmethod
    def add_cat_labels(cls, chart_doc: XChartDocument, data_label: str, data_range: str) -> None:
        """
        Add Category Labels

        Args:
            chart_doc (XChartDocument): Chart Document
            data_label (str): Data label
            data_range (str): Data range

        Raises:
            ChartError: If error occurs.

        Returns:
            None:
        """
        try:
            dp = chart_doc.getDataProvider()
            dl_seq = cls.create_ld_seq(
                dp=dp, role=DataRoleKind.CATEGORIES, data_label=data_label, data_range=data_range
            )

            axis = cls.get_axis(chart_doc=chart_doc, axis_val=AxisKind.X, idx=0)
            sd = axis.getScaleData()
            sd.Categories = dl_seq
            axis.setScaleData(sd)

            # abel the data points with these category values
            cls.set_data_point_labels(chart_doc=chart_doc, label_type=DataPointLabelTypeKind.CATEGORY)
        except mEx.ChartError:
            raise
        except Exception as e:
            raise mEx.ChartError("Error adding category lables") from e

    # endregion add data to a chart

    # region chart shape and image
    @staticmethod
    def get_chart_shape(sheet: XSpreadsheet) -> XShape:
        """
        Gets chart shape

        Args:
            sheet (XSpreadsheet): Spreadsheet

        Raises:
            ShapeMissingError: If shape is ``None``.
            ShapeError: If any other error occurs.

        Returns:
            XShape: Shape object.
        """
        shape = None
        try:
            page_supp = mLo.Lo.qi(XDrawPageSupplier, sheet, True)
            draw_page = page_supp.getDrawPage()
            num_shapes = draw_page.getCount()
            chart_classid = mLo.Lo.CLSID.CHART.value
            for i in range(num_shapes):
                try:
                    shape = mLo.Lo.qi(XShape, draw_page.getByIndex(i), True)
                    classid = str(mProps.Props.get_property(shape, "CLSID")).lower()
                    if classid == chart_classid:
                        break
                except Exception:
                    shape = None
                    # continue on, just because got an error does not mean shape will not be found
        except Exception as e:
            raise mEx.ShapeError("Error getting shape from sheet") from e
        if shape is None:
            raise mEx.ShapeMissingError("Unalbe to find Chart Shape")
        return shape

    @classmethod
    def copy_chart(cls, ssdoc: XSpreadsheetDocument, sheet: XSpreadsheet) -> None:
        """
        Copies a chart using a dispatch command.

        Args:
            ssdoc (XSpreadsheetDocument): Spreadsheet Document
            sheet (XSpreadsheet): Spreadsheet

        Raises:
            ChartError: If error occurs.

        Returns:
            None:
        """
        try:
            chart_shape = cls.get_chart_shape(sheet=sheet)
            doc = mLo.Lo.qi(XComponent, ssdoc, True)
            supp = mGui.GUI.get_selection_supplier(doc)
            supp.select(chart_shape)
            mLo.Lo.dispatch_cmd("Copy")
        except Exception as e:
            raise mEx.ChartError("Error in attempt to copy chart") from e

    @classmethod
    def get_chart_draw_page(cls, sheet: XSpreadsheet) -> XDrawPage:
        """
        Gets chart draw page.

        Args:
            sheet (XSpreadsheet): Spreadsheet

        Raises:
            ChartError: If error occurs.

        Returns:
            XDrawPage: Draw Page object
        """
        try:
            chart_shape = cls.get_chart_shape(sheet)
            embedded_chart = mLo.Lo.qi(XEmbeddedObject, mProps.Props.get_property(chart_shape, "EmbeddedObject"), True)
            comp_supp = mLo.Lo.qi(XComponentSupplier, embedded_chart, True)
            xclosable = comp_supp.getComponent()
            supp_page = mLo.Lo.qi(XDrawPageSupplier, xclosable, True)
            result = supp_page.getDrawPage()
            if result is None:
                raise mEx.UnKnownError("None Value: getDrawPage() returned a value of None")
            return result
        except Exception as e:
            raise mEx.ChartError("Error getting chart draw page") from e

    @classmethod
    def get_chart_image(cls, sheet: XSpreadsheet) -> XGraphic:
        """
        Get chart image as ``XGraphic``.

        Args:
            sheet (XSpreadsheet): Spreadsheet

        Raises:
            ChartError: If error occurs.

        Returns:
            XGraphic: Graphic object
        """
        try:
            chart_shape = cls.get_chart_shape(sheet)

            graphic = mLo.Lo.qi(XGraphic, mProps.Props.get_property(chart_shape, "Graphic"), True)

            tmp_fnm = mFileIo.FileIO.create_temp_file("png")
            mImgLo.ImagesLo.save_graphic(pic=graphic, fnm=tmp_fnm, im_format="png")
            im = mImgLo.ImagesLo.load_graphic_file(tmp_fnm)
            mFileIo.FileIO.delete_file(tmp_fnm)
            return im
        except Exception as e:
            raise mEx.ChartError("Error getting chart image") from e

    # endregion chart shape and image


__all__ = ("Chart2",)



from enum import Enum
from . import kind_helper


class DataRoleKind(str, Enum):
    """
    Represents DataSequenceRole

    See Also:
        `DataSequenceRole API <https://api.libreoffice.org/docs/idl/ref/namespacecom_1_1sun_1_1star_1_1chart2_1_1data.html#a340775895b509a7f80ba895767123429>`_
    """

    CATEGORIES = "categories"
    """Values are used for categories in the diagram"""
    ERROR_BARS_X_NEGATIVE = "error-bars-x-negative"
    """Values are used as error-information in negative x-direction for displaying error-bars"""
    ERROR_BARS_X_POSITIVE = "error-bars-x-positive"
    """Values are used as error-information in positive x-direction for displaying error-bars"""
    ERROR_BARS_Y_NEGATIVE = "error-bars-y-negative"
    """Values are used as error-information in negative y-direction for displaying error-bars"""
    ERROR_BARS_Y_POSITIVE = "error-bars-y-positive"
    """Values are used as error-information in positive y-direction for displaying error-bars"""
    LABEL = "label"
    """Values are used as a label for a series. Usually, you will have just one cell containing a string."""
    SIZES = "sizes"
    """values are used as radius of the bubbles in a Bubble-Diagram."""
    VALUES_FIRST = "values-first"
    """``Candle-stick`` chart - the first value of a series of values. In a stock-chart this would be the opening course."""
    VALUES_LAST = "values-last"
    """``Candle-stick`` chart - the last value of a series of values. In a stock-chart this would be the closing course."""
    VALUES_MAX = "values-max"
    """``Candle-stick`` chart - the maximum value of a series of values. In a stock-chart this would be the highest course that occurred during trading."""
    VALUES_MIN = "values-min"
    """``Candle-stick`` chart - the minimum value of a series of values. In a stock-chart this would be the lowest course that occurred during trading."""
    VALUES_X = "values-x"
    """Values are used as x-values in an ``XY`` - or bubble diagram."""
    VALUES_Y = "values-y"
    """Values are used as y-values in an ``XY-Diagram`` or as values in a bar, line, etc. chart."""
    VALUES_Z = "values-z"
    """Values may be used as z-values in a three-dimensional ``XYZ-Diagram`` or a surface-chart."""

    def __str__(self) -> str:
        return str(self.value)

    @staticmethod
    def from_str(s: str) -> "DataRoleKind":
        """
        Gets an ``DataRoleKind`` instance from string.

        Args:
            s (str): String that represents the name of an enum Name.
                ``s`` is case insensitive and can be ``CamelCase``, ``pascal_case`` , ``snake_case``,
                ``hypen-case``, ``normal case``.

        Raises:
            ValueError: If input string is empty.
            AttributeError: If unable to get ``DataRoleKind`` instance.

        Returns:
            DataRoleKind: Enum instance.
        """
        return kind_helper.enum_from_string(s, DataRoleKind)



# coding: utf-8
from __future__ import annotations
import os
from pathlib import Path
from shutil import copytree
import shutil
import tempfile
from ..utils.type_var import PathOrStr
from ..utils import sys_info
from ..cfg import config


class Cache:
    """Office Profile Cache Manager"""

    def __init__(self, **kwargs) -> None:
        """
        Cache Constructor

        Keyword Args:
            use_cache (bool, optional): Determines if caching is used. Default is True
            cache_path (PathOrStr, optional): Set the path used for caching LO profile data.
                Default is searched for in known locations.
            working_dir (PathOrStr, optional): sets the working dir to use.
                This is the dir that LO profile will be copied to. Defaults to a newly generated temp dir.
        """
        self._use_cache = bool(kwargs.get("use_cache", True))
        self._profile_dir_name = "profile"
        self._profile_cached = False
        cache_path = kwargs.get("cache_path", None)
        if cache_path is not None:
            self.cache_path = cache_path
        working_dir = kwargs.get("working_dir", None)
        if working_dir is not None:
            self.working_dir = working_dir

    def _get_cache_path(self) -> Path | None:
        # this method is only ever called the user does not provide a cache_path
        # see: https://www.howtogeek.com/289587/how-to-find-your-libreoffice-profile-folder-in-windows-macos-and-linux/
        cache_path = None
        platform = sys_info.SysInfo.get_platform()
        def get_path(ver: str):
            result = None
            if platform == sys_info.SysInfo.PlatformEnum.LINUX:
                result = Path("~/.config/libreoffice", ver).expanduser()
            elif platform == sys_info.SysInfo.PlatformEnum.WINDOWS:
                result = Path(os.getenv("APPDATA"), "LibreOffice", ver)
            elif platform == sys_info.SysInfo.PlatformEnum.MAC:
                result = Path("~/Library/Application Support/LibreOffice/", ver).expanduser()
            if result is not None:
                if result.exists() is False or result.is_dir() is False:
                    result = None
            if result is None:
                return False, None
            return True, result

        # lookup profile versions from config
        for s_ver in config.Config().profile_versions:
            is_valid, cache_path = get_path(s_ver)
            if is_valid:
                break

        return cache_path

    def cache_profile(self) -> None:
        """
        Copies user profile into cache path if it has not already been cached.

        Ignored if :py:attr:`~Cache.use_cache` is ``False``
        """
        # copy_cache_to_profile is called before this method
        if self.use_cache is False:
            return
        if self.cache_path is None:
            return
        if self._profile_cached is False:
            copytree(self.user_profile, self.cache_path)
        return

    def copy_cache_to_profile(self) -> None:
        """
        Copies cached profile to profile dir set in
        :py:attr:`~.Cache.user_profile`

        Ignored if :py:attr:`~Cache.use_cache` is ``False``
        """
        # this method is called before cache_profile.
        if self.use_cache is False:
            return
        if self.cache_path is None:
            return
        if self.cache_path.exists() and self.cache_path.is_dir():
            copytree(self.cache_path, self.user_profile)
            self._profile_cached = True
        else:
            # create the dir.
            # when when cache_profile is called it will cache
            # the profile into this dir.
            os.mkdir(self.user_profile)
            self._profile_cached = False

    def del_working_dir(self):
        """
        Deletes the current working directory of instance.

        Ignored if :py:attr:`~Cache.use_cache` is ``False``
        """
        if self.use_cache:
            if self.working_dir.exists() and self.working_dir.is_dir():
                shutil.rmtree(self.working_dir)

    @property
    def user_profile(self) -> Path:
        """Gets user profile path"""
        try:
            return self._user_profile
        except AttributeError:
            self._user_profile = Path(self.working_dir, self._profile_dir_name)
        return self._user_profile

    @property
    def cache_path(self) -> Path | None:
        """
        Gets/Sets cache path

        The when possible default is obtained by searching in know locations, depending on OS.
        For instance on Linux will search in ``~/.config/libreoffice/4``
        """
        try:
            return self._cache_path
        except AttributeError:
            self._cache_path = self._get_cache_path()
        return self._cache_path

    @cache_path.setter
    def cache_path(self, value: PathOrStr | None):
        self._cache_path = Path(value)

    @property
    def working_dir(self) -> Path:
        """
        Gets/Sets working dir

        This property determines the dir LO uses to obtain its profile data.
        Default to an auto generated tmp dir.
        """
        try:
            return self._working_dir
        except AttributeError:
            self._working_dir = Path(tempfile.mkdtemp())
        return self._working_dir

    @working_dir.setter
    def working_dir(self, value: PathOrStr):
        self._working_dir = Path(value)

    @property
    def use_cache(self) -> bool:
        """Gets/Sets if cache is used. Default is ``True``"""
        return self._use_cache

    @use_cache.setter
    def use_cache(self, value: bool) -> None:
        self._use_cache = value



# coding: utf-8
from __future__ import annotations, unicode_literals
import sys
from typing import TYPE_CHECKING, cast
import uno
import getpass, os, os.path
from ..meta.static_meta import StaticProperty, classproperty
from ..events.args.event_args import EventArgs
from ..events.lo_named_event import LoNamedEvent
from ..events.event_singleton import _Events
from . import lo as mLo
from ..exceptions import ex as mEx

from com.sun.star.util import XStringSubstitution
from ooo.dyn.uno.deployment_exception import DeploymentException

# com.sun.star.uno.DeploymentException

if TYPE_CHECKING:
    from com.sun.star.util import PathSubstitution
    from com.sun.star.uno import XComponentContext


class Session(metaclass=StaticProperty):
    """
    Session Class for handling user paths within the LibreOffice environment.

    See Also:
        - `Importing Python Modules <https://help.libreoffice.org/latest/lo/text/sbasic/python/python_import.html>`_
        - `Getting Session Information <https://help.libreoffice.org/latest/lo/text/sbasic/python/python_session.html>`_
    """

    # https://help.libreoffice.org/latest/lo/text/sbasic/python/python_import.html
    # https://help.libreoffice.org/latest/lo/text/sbasic/python/python_session.html

    class PathEnum:
        SHARE_PYTHON = 1
        SHARE_USER_PYTHON = 2

    @classproperty
    def path_sub(cls) -> PathSubstitution:
        try:
            return cls._path_substitution
        except AttributeError:
            try:
                # raises DeploymentException if not in a macro environment
                # in a macro envrionment there should be no dependency on Lo
                # this will allow for import shared python files before Lo.load_office is called.
                # if not in a macro then must get instace after Lo.load_office is called
                ctx = cast("XComponentContext", uno.getComponentContext())
                ps = ctx.getServiceManager().createInstanceWithContext("com.sun.star.util.PathSubstitution", ctx)
                cls._path_substitution = ps
                return cls._path_substitution
            except DeploymentException as e:
                # print(e)
                # there must be a connection to before calling session.
                if mLo.Lo.is_loaded is False:
                    raise mEx.LoNotLoadedError(
                        "Lo.load_office must be called before using session when not run as a macro"
                    )
                cls._path_substitution = mLo.Lo.create_instance_mcf(
                    XStringSubstitution, "com.sun.star.util.PathSubstitution"
                )
        return cls._path_substitution

    @staticmethod
    def substitute(var_name: str):
        """
        Returns the current value of a variable.

        The method iterates through its internal variable list and tries to find the given variable.
        If the variable is unknown a com.sun.star.container.NoSuchElementException is thrown.

        Args:
            var_name (str): name to search for.

        Raises:
            com.sun.star.container.NoSuchElementException: ``NoSuchElementException``
        """
        return Session.path_sub.getSubstituteVariableValue(var_name)

    @classproperty
    def share(cls) -> str:
        """
        Gets Program Share dir,
        such as ``C:\\Program Files\\LibreOffice\\share``
        """
        try:
            return cls._share
        except AttributeError:
            inst = uno.fileUrlToSystemPath(Session.substitute("$(prog)"))
            cls._share = os.path.normpath(inst.replace("program", "Share"))
        return cls._share

    @classproperty
    def shared_scripts(cls) -> str:
        """
        Gets Program Share scripts dir,
        such as ``C:\\Program Files\\LibreOffice\\share\\Scripts``
        """
        return "".join([Session.share, os.sep, "Scripts"])

    @classproperty
    def shared_py_scripts(cls) -> str:
        """
        Gets Program Share python dir,
        such as ``C:\\Program Files\\LibreOffice\\share\\Scripts\\python``
        """
        # eg: C:\Program Files\LibreOffice\share\Scripts\python
        return "".join([Session.shared_scripts, os.sep, "python"])

    @classproperty  # alternative to '$(username)' variable
    def user_name(cls):
        """Get the username from the environment or password database.

        First try various environment variables, then the password
        database.  This works on Windows as long as USERNAME is set.
        """
        return getpass.getuser()

    @classproperty
    def user_profile(cls):
        """
        Gets path to user profile such as,
        ``C:\\Users\\user\\AppData\\Roaming\\LibreOffice\\4\\user``
        """
        try:
            return cls._user_profile
        except AttributeError:
            cls._user_profile = uno.fileUrlToSystemPath(Session.substitute("$(user)"))
        return cls._user_profile

    @classproperty
    def user_scripts(cls):
        """
        Gets path to user profile scripts such as,
        ``C:\\Users\\user\\AppData\\Roaming\\LibreOffice\\4\\user\\Scripts``
        """
        return "".join([cls.user_profile, os.sep, "Scripts"])

    @classproperty
    def user_py_scripts(cls):
        """
        Gets path to user profile python such as,
        ``C:\\Users\\user\\AppData\\Roaming\\LibreOffice\\4\\user\\Scripts\\python``
        """
        # eg: C:\Users\user\AppData\Roaming\LibreOffice\4\user\Scripts\python
        return "".join([cls.user_scripts, os.sep, "python"])

    @classmethod
    def register_path(cls, path: Session.PathEnum) -> None:
        """
        Registers a path into ``sys.path`` if it does not exist

        Args:
            path (PathEnum): Type of path to register.
        """
        if path == Session.PathEnum.SHARE_PYTHON:
            spath = cls.shared_py_scripts
        elif path == Session.PathEnum.SHARE_USER_PYTHON:
            spath = cls.user_py_scripts
        if not spath in sys.path:
            sys.path.insert(0, spath)


def _del_cache_attrs(source: object, e: EventArgs) -> None:
    # clears Lo Attributes that are dynamically created
    dattrs = ("_path_substitution", "_share", "_user_profile")
    for attr in dattrs:
        if hasattr(Session, attr):
            delattr(Session, attr)


_Events().on(LoNamedEvent.RESET, _del_cache_attrs)

__all__ = ("Session",)



# coding: utf-8
"""
This module is for the purpose of sharing events between classes.
"""
from __future__ import annotations
import contextlib
from weakref import ref, ReferenceType, proxy
from typing import Any, Dict, List, NamedTuple, Generator
from .args.event_args import EventArgs

from ..utils.type_var import EventCallback as EventCallback
from ..proto import event_observer
from . import event_singleton


class EventArg(NamedTuple):
    """
    Event Args for passing event to :py:func:`~.event_ctx`
    """

    name: str
    """Event name"""
    callback: EventCallback
    """
    Event Callback
    """


class _event_base(object):
    """Base events class"""

    def __init__(self) -> None:
        self._callbacks = None

    def on(self, event_name: str, callback: EventCallback):
        """
        Registers an event

        Args:
            event_name (str): Unique event name
            callback (Callable[[object, EventArgs], None]): Callback function
        """
        if self._callbacks is None:
            self._callbacks: Dict[str, List[ReferenceType[EventCallback]]] = {}

        if event_name not in self._callbacks:
            self._callbacks[event_name] = [ref(callback)]
        else:
            self._callbacks[event_name].append(ref(callback))

    def remove(self, event_name: str, callback: EventCallback) -> bool:
        """
        Removes an event callback

        Args:
            event_name (str): Unique event name
            callback (Callable[[object, EventArgs], None]): Callback function

        Returns:
            bool: True if callback has been removed; Otherwise, False.
            False means the callback was not found.
        """
        if self._callbacks is None:
            return False
        result = False
        if event_name in self._callbacks:
            # cb = cast(Dict[str, List[EventCallback]], self._callbacks)
            try:
                self._callbacks[event_name].remove(ref(callback))
                result = True
            except ValueError:
                pass
        return result

    def _set_event_args(self, event_name: str, event_args: EventArgs) -> None:
        if event_args is None:
            return
        event_args._event_name = event_name
        event_args._event_source = self

    def trigger(self, event_name: str, event_args: EventArgs, *args, **kwargs):
        """
        Trigger event(s) for a given name.

        Args:
            event_name (str): Name of event to trigger
            event_args (EventArgs): Event args passed to the callback for trigger.
            args (Any, optional): Optional positional args to pass to callback
            kwargs (Any, optional): Optional keyword args to pass to callback

        Note:
            Events are removed automatically when they are out of scope.
        """

        if self._callbacks is not None and event_name in self._callbacks:
            cleanup = None
            for i, callback in enumerate(self._callbacks[event_name]):
                if callback() is None:
                    if cleanup is None:
                        cleanup = []
                    cleanup.append(i)
                    continue
                self._set_event_args(event_name=event_name, event_args=event_args)
                if callable(callback()):
                    try:
                        callback()(event_args.source, event_args, *args, **kwargs)
                    except AttributeError:
                        # event_arg is None
                        callback()(self, None)
            if cleanup is not None and len(cleanup) > 0:
                cleanup.reverse()
                for i in cleanup:
                    self._callbacks[event_name].pop(i)
                if len(self._callbacks[event_name]) == 0:
                    del self._callbacks[event_name]


class Events(_event_base):
    """Class for sharing events among classes and functions."""

    # Dev Notes:
    # Event callbacks are assigned to this class as a weak ref.
    # This is necessary; Howerver, a side effect is class method cannot be assigned
    # as an event from class __init__. It is possible to assign a class static method from
    # __init__ but not a class method. Attempting to assign class method result with method
    # being out of scope before trigger is called on it.
    # Making an Events class with strong ref ( no weak ref ) and then assigning a class method
    # as a callback result in the class method being triggered even after the class instance is set
    # to none. In other words python does not realease the object or callback because the strong ref Events class
    # is still holding on to it.
    # In short, do not change this class!

    def __init__(self, source: Any | None = None) -> None:
        """
        Construct for Events

        Args:
            source (Any | None, optional): Source can be class or any object.
                The value of ``source`` is the value assigned to the ``EventArgs.event_source`` property.
                Defaults to current instance of this class.
        """
        super().__init__()
        self._source = source
        # register wih LoEvents so this instance get triggered when LoEvents() are triggered.
        LoEvents().add_observer(self)


    def _set_event_args(self, event_name: str, event_args: EventArgs) -> None:
        if event_args is None:
            return
        event_args._event_name = event_name
        event_args._event_source = self if self._source is None else self._source

class LoEvents(_event_base):
    """Singleton Class for ODEV global events."""

    _instance = None

    def __new__(cls, *args, **kwargs):
        if not cls._instance:
            cls._instance = super(LoEvents, cls).__new__(cls, *args, **kwargs)
            cls._instance._callbacks = None
            cls._instance._observers: List[ReferenceType[event_observer.EventObserver]] = None
            # register wih _Events so this instance get triggered when _Events() are triggered.
            event_singleton._Events().add_observer(cls._instance)
        return cls._instance

    def __init__(self) -> None:
        pass

    def add_observer(self, *args: event_observer.EventObserver) -> None:
        """
        Adds observers that gets their ``trigger`` method called when this class ``trigger`` method is called.

        Parameters:
            args (EventObserver): One or more observers to add.

        Returns:
            None:

        Note:
            Observers are removed automatically when they are out of scope.
        """
        if self._observers is None:
            self._observers = []
        for observer in args:
            self._observers.append(ref(observer))

    def trigger(self, event_name: str, event_args: EventArgs):
        super().trigger(event_name, event_args)
        self._update_observers(event_name, event_args)

    def _update_observers(self, event_name: str, event_args: EventArgs) -> None:
        if self._observers is not None:
            cleanup = None
            for i, observer in enumerate(self._observers):
                if observer() is None:
                    if cleanup is None:
                        cleanup = []
                    cleanup.append(i)
                    continue
                observer().trigger(event_name=event_name, event_args=event_args)
            if cleanup is not None and len(cleanup) > 0:
                # reverse list to allow removing form highest to lowest to avoid errors
                cleanup.reverse()
                for i in cleanup:
                    self._observers.pop(i)


@contextlib.contextmanager
def event_ctx(*args: EventArg) -> Generator[event_observer.EventObserver, None, None]:
    """
    Event context manager.

    This manager adds and removes events.

    Parameters:
        args (EventArg): One or more EventArgs to add.

    Yields:
        Generator[EventObserver, None, None]: events
    """
    try:
        # yields a weakref.proxy obj
        # wekaref is dead as soon as e_obj is set to none.
        e_obj = Events()  # automatically adds itself as an observer to LoEvents()
        for arg in args:
            e_obj.on(arg.name, arg.callback)
        yield proxy(e_obj)
    except Exception:
        raise
    finally:
        e_obj = None
        _ = None  # just to make sure _ is not a ref to e_obj


def is_meth_event(source: str, meth: callable) -> bool:
    """
    Gets if event source is the same as meth.
    This method for for core events.

    Args:
        source (str): source as str
        meth (callable): method to test.

    Returns:
        bool: True if event is raised by meth; Otherwise; False
    """
    try:
        return source == meth.__qualname__
    except Exception:
        pass
    return False



# coding: utf-8
from __future__ import annotations
from typing import Any, List, TYPE_CHECKING
from ..utils.type_var import PathOrStr

if TYPE_CHECKING:
    from ..events.args.event_args import EventArgs


class NoneError(Exception):
    """Generic None Error. Usually means object is None"""

    pass


class NotFoundError(Exception):
    """Generic Not Found Error"""

    pass


class MissingInterfaceError(NotFoundError):
    """Error when a interface is not found for a uno object"""

    def __init__(self, interface: Any, message: Any = None, *args) -> None:
        """
        MissingInterfaceError constructor

        Args:
            interface (Any): Missing Interface that caused error
            message (Any, optional): Message of error
        """
        if message is None:
            try:
                message = f"Missing interface {interface.__pyunointerface__}"
            except AttributeError:
                message = "Missing Uno Interface Error"
        super().__init__(interface, message, *args)

    def __str__(self) -> str:
        return repr(self.args[1])


class CellError(Exception):
    """Cell error"""

    pass


class ConfigError(Exception):
    """Config Error"""

    pass


class PropertyGeneralError(Exception):
    """Generic Property Error"""

    pass


class PropertySetError(PropertyGeneralError):
    """Generic Property Set Error"""

    pass


class PropertySetMissingError(PropertySetError):
    """Error for missing property set"""

    pass


class PropertyError(PropertyGeneralError):
    """
    Property Error
    """

    def __init__(self, prop_name: str, *args: object) -> None:
        """
        PropertyError Constructor

        Args:
            prop_name (str): Property name that caused error
        """
        super().__init__(prop_name, *args)

    def __str__(self) -> str:
        return repr(f"Property Error for: {self.args[0]}")


class PropertiesError(PropertyGeneralError):
    """Error for multiple properties"""

    pass


class PropertyNotFoundError(PropertyError):
    """Property Not Found Error"""

    def __str__(self) -> str:
        return repr(f"Property not found for: {self.args[0]}")


class GoalDivergenceError(Exception):
    """Error when goal seek result divergence is too high"""

    def __init__(self, divergence: float, message: Any = None) -> None:
        """
        GoalDivergenceError Constructor

        Args:
            divergence (float): divergence amount
            message (Any, optional): Message of error
        """
        if message is None:
            message = f"Divergence error: {divergence:.4f}"
        super().__init__(divergence, message)

    def __str__(self) -> str:
        return repr(self.args[1])


class UnKnownError(Exception):
    """Error for unknown results"""

    pass


class UnOpenableError(Exception):
    def __init__(self, fnm: PathOrStr, *args: object) -> None:
        """
        PropertyError Constructor

        Args:
            fnm (PathOrStr): File path that is not able to be opened.
        """
        super().__init__(fnm, *args)

    def __str__(self) -> str:
        return repr(f"Un-openable file: '{self.args[0]}'")


class MultiError(Exception):
    """Handles Multiple errors"""

    def __init__(self, errors: List[Exception]) -> None:
        """
        MultiError Constructor

        Args:
            errors (List[Exception]): List of errors
        """
        self.errors = errors
        super().__init__(self.errors)

    def __str__(self) -> str:
        return "\n".join([str(x) for x in self.errors])


class NotSupportedError(Exception):
    """Generic Not Supported Error"""

    pass


class NotSupportedServiceError(NotSupportedError):
    """
    Handles errors of service not being supported.
    """

    def __init__(self, service_name: str, *args: object) -> None:
        """
        NotSupportedServiceError Constructor

        Args:
            service_name (str): Service name
        """
        super().__init__(service_name, *args)

    def __str__(self) -> str:
        return repr(f"Service not supported: '{self.args[0]}'")


class NotSupportedMacroModeError(NotSupportedError):
    """
    Handles errors of operations that are not allow when running as a macro.

    This error is largely used for methods that require external imports
    such as XML.apply_xslt()
    """

    pass


class CreateInstanceError(Exception):
    """Create instance Error"""

    def __init__(self, interface: Any, service: str, message: Any = None, *args) -> None:
        """
        constructor

        Args:
            interface (Any): Interface that failed creation
            message (Any, optional): Message of error
        """
        if message is None:
            message = f"Unable to create instance of {service}"
        super().__init__(interface, service, message, *args)

    def __str__(self) -> str:
        try:
            interface_name = self.args[0].__pyunointerface__
        except AttributeError:
            interface_name = "Unknown Interface"
        return f"Unable to create instance for service '{self.args[1]}' with interface of '{interface_name}'.\n{self.args[2]}"


class CreateInstanceMsfError(CreateInstanceError):
    """Create MSF Instance Error"""

    pass


class CreateInstanceMcfError(CreateInstanceError):
    """Create MCF Instance Error"""

    pass


class CancelEventError(Exception):
    """Error when an Event is canceled"""

    def __init__(self, event_args: EventArgs, message: Any = None, *args) -> None:
        """
        Cancel Event Error constructor

        Args:
            event_args (EventArgs): Event args that was canceled
            message (Any, optional): Message of error
        """
        if message is None:
            message = f"Event '{event_args.event_name}' is canceled!"
        super().__init__(event_args, message, *args)

    def __str__(self) -> str:
        return repr(self.args[1])


class CursorError(Exception):
    """Handles Cursor errors"""

    pass


class WordCursorError(CursorError):
    """Handles Word Cursor errors"""

    pass


class LineCursorError(CursorError):
    """Handles Line Cursor errors"""

    pass


class SentenceCursorError(CursorError):
    """Handles Sentence Cursor errors"""

    pass


class ParagraphCursorError(CursorError):
    """Handles Sentence Cursor errors"""

    pass


class PageCursorError(Exception):
    """Handles Page Cursor errors"""

    pass


class ViewCursorError(CursorError):
    """Handles View Cursor errors"""

    pass


class LoNotLoadedError(Exception):
    """Error when accessing Lo before Lo.load_office is called"""

    pass


class ChartError(Exception):
    """Generic Chart Error"""

    pass


class ChartNoAccessError(ChartError):
    """No Access to Chart"""

    pass


class ChartExistingError(ChartError):
    """Chart already exist Error"""

    pass


class ChartNotExistingError(ChartError):
    """Chart does not exist"""

    pass


class DiagramError(Exception):
    """Generic Diagram Error"""

    pass


class DiagramNotExistingError(DiagramError):
    """Diagram  does not exist Error"""

    pass


class DrawError(Exception):
    """Generic Draw Error"""

    pass


class DrawPageError(DrawError):
    """Generic Draw Page Error"""

    pass


class DrawPageMissingError(DrawPageError):
    """Draw page Missing Error"""

    pass


class ShapeError(Exception):
    """Generic Shape Error"""

    pass


class ShapeMissingError(ShapeError):
    """Missing Shape Error"""

    pass


class SizeError(Exception):
    """Generic Size Error"""

    pass


class PointError(Exception):
    """Generic Point Error"""

    pass


class ColorError(Exception):
    """Generic Color Error"""

    pass


class ServiceError(Exception):
    """Generic Service Error"""

    pass


class ServiceNotSupported(ServiceError):
    """
    Service not supported error
    """

    def __init__(self, *service: str, message="") -> None:
        """
        Constructor

        Args:
            message (str, optional): Extra error message.
            *service (str): Variable length argument list of UNO namespace strings such as ``com.sun.star.configuration.GroupAccess``
        """
        self.service = service
        self.message = message
        super().__init__(self.message)

    def __str__(self) -> str:
        msg = ""
        serv_len = len(self.service)
        if serv_len == 1:
            msg = f"Service not supported for: {self.service[0]}"
        elif serv_len > 1:
            msg = "Services not supported for: "
            msg += ", ".join(self.service)

        if self.message:
            return f"{self.message} -> {msg}"

        return msg

    def __repr__(self) -> str:
        service_str = ", ".join([f'"{s}"' for s in self.service])
        return f'ServiceNotSupported({service_str}, message="{self.message}")'


class GalleryError(Exception):
    """Generic Gallery Error"""

    pass


class GalleryTypeError(GalleryError):
    """
    Error when wrong gallery type is used.

    Such as when trying to extract a ``XGraphic`` from a ``XGalleryItem`` that is not a graphic.
    """

    pass


class GalleryNotFoundError(GalleryError):
    """Error when Gallery Item is not found"""

    pass


class ImageError(Exception):
    """Generic error when error occurs processing an image"""

    pass


class DispatchError(Exception):
    """Generic error when an error occurs while dispatching"""

    pass
